@@ -3,45 +3,29 @@
  * sequence.c
  *	  PostgreSQL sequences support code.
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/commands/sequence.c,v 1.68.2.1 2002/03/15 19:20:45 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/sequence.c,v 1.119 2004/12/31 21:59:41 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include <ctype.h>
-
 #include "access/heapam.h"
-#include "commands/creatinh.h"
+#include "catalog/namespace.h"
+#include "catalog/pg_type.h"
+#include "commands/defrem.h"
+#include "commands/tablecmds.h"
 #include "commands/sequence.h"
 #include "miscadmin.h"
 #include "utils/acl.h"
 #include "utils/builtins.h"
-#include "utils/int8.h"
-#ifdef MULTIBYTE
-#include "mb/pg_wchar.h"
-#endif
+#include "utils/resowner.h"
 
 
-#define SEQ_MAGIC	  0x1717
-
-#ifndef INT64_IS_BUSTED
-#ifdef HAVE_LL_CONSTANTS
-#define SEQ_MAXVALUE	((int64) 0x7FFFFFFFFFFFFFFFLL)
-#else
-#define SEQ_MAXVALUE	((int64) 0x7FFFFFFFFFFFFFFF)
-#endif
-#else							/* INT64_IS_BUSTED */
-#define SEQ_MAXVALUE	((int64) 0x7FFFFFFF)
-#endif   /* INT64_IS_BUSTED */
-
-#define SEQ_MINVALUE	(-SEQ_MAXVALUE)
-
 /*
  * We don't want to log each fetching of a value from a sequence,
  * so we pre-log a few fetches in advance. In the event of
@@ -49,32 +33,47 @@
  */
 #define SEQ_LOG_VALS	32
 
+/*
+ * The "special area" of a sequence's buffer page looks like this.
+ */
+#define SEQ_MAGIC	  0x1717
+
 typedef struct sequence_magic
 {
 	uint32		magic;
 } sequence_magic;
 
+/*
+ * We store a SeqTable item for every sequence we have touched in the current
+ * session.  This is needed to hold onto nextval/currval state.  (We can't
+ * rely on the relcache, since it's only, well, a cache, and may decide to
+ * discard entries.)
+ *
+ * XXX We use linear search to find pre-existing SeqTable entries.	This is
+ * good when only a small number of sequences are touched in a session, but
+ * would suck with many different sequences.  Perhaps use a hashtable someday.
+ */
 typedef struct SeqTableData
 {
-	char	   *name;
-	Oid			relid;
-	Relation	rel;			/* NULL if rel is not open in cur xact */
-	int64		cached;
-	int64		last;
-	int64		increment;
-	struct SeqTableData *next;
+	struct SeqTableData *next;	/* link to next SeqTable object */
+	Oid			relid;			/* pg_class OID of this sequence */
+	TransactionId xid;			/* xact in which we last did a seq op */
+	int64		last;			/* value last returned by nextval */
+	int64		cached;			/* last value already cached for nextval */
+	/* if last != cached, we have not used up all the cached values */
+	int64		increment;		/* copy of sequence's increment field */
 } SeqTableData;
 
 typedef SeqTableData *SeqTable;
 
-static SeqTable seqtab = NULL;
+static SeqTable seqtab = NULL;	/* Head of list of SeqTable items */
 
-static char *get_seq_name(text *seqin);
-static SeqTable init_sequence(char *caller, char *name);
-static Form_pg_sequence read_info(char *caller, SeqTable elm, Buffer *buf);
-static void init_params(CreateSeqStmt *seq, Form_pg_sequence new);
-static int64 get_param(DefElem *def);
-static void do_setval(char *seqname, int64 next, bool iscalled);
+
+static void init_sequence(RangeVar *relation,
+			  SeqTable *p_elm, Relation *p_rel);
+static Form_pg_sequence read_info(SeqTable elm, Relation rel, Buffer *buf);
+static void init_params(List *options, Form_pg_sequence new, bool isInit);
+static void do_setval(RangeVar *sequence, int64 next, bool iscalled);
 
 /*
  * DefineSequence
@@ -85,8 +84,7 @@ DefineSequence(CreateSeqStmt *seq)
 {
 	FormData_pg_sequence new;
 	CreateStmt *stmt = makeNode(CreateStmt);
-	ColumnDef  *coldef;
-	TypeName   *typnam;
+	Oid			seqoid;
 	Relation	rel;
 	Buffer		buf;
 	PageHeader	page;
@@ -98,8 +96,8 @@ DefineSequence(CreateSeqStmt *seq)
 	int			i;
 	NameData	name;
 
-	/* Check and set values */
-	init_params(seq, &new);
+	/* Check and set all option values */
+	init_params(seq->options, &new, true);
 
 	/*
 	 * Create relation (and fill *null & *value)
@@ -107,62 +105,71 @@ DefineSequence(CreateSeqStmt *seq)
 	stmt->tableElts = NIL;
 	for (i = SEQ_COL_FIRSTCOL; i <= SEQ_COL_LASTCOL; i++)
 	{
+		ColumnDef  *coldef;
+		TypeName   *typnam;
+
 		typnam = makeNode(TypeName);
 		typnam->setof = FALSE;
-		typnam->arrayBounds = NULL;
+		typnam->arrayBounds = NIL;
 		typnam->typmod = -1;
+
 		coldef = makeNode(ColumnDef);
 		coldef->typename = typnam;
+		coldef->inhcount = 0;
+		coldef->is_local = true;
+		coldef->is_not_null = true;
 		coldef->raw_default = NULL;
 		coldef->cooked_default = NULL;
-		coldef->is_not_null = false;
+		coldef->constraints = NIL;
+		coldef->support = NULL;
+
 		null[i - 1] = ' ';
 
 		switch (i)
 		{
 			case SEQ_COL_NAME:
-				typnam->name = "name";
+				typnam->typeid = NAMEOID;
 				coldef->colname = "sequence_name";
-				namestrcpy(&name, seq->seqname);
+				namestrcpy(&name, seq->sequence->relname);
 				value[i - 1] = NameGetDatum(&name);
 				break;
 			case SEQ_COL_LASTVAL:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "last_value";
 				value[i - 1] = Int64GetDatumFast(new.last_value);
 				break;
 			case SEQ_COL_INCBY:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "increment_by";
 				value[i - 1] = Int64GetDatumFast(new.increment_by);
 				break;
 			case SEQ_COL_MAXVALUE:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "max_value";
 				value[i - 1] = Int64GetDatumFast(new.max_value);
 				break;
 			case SEQ_COL_MINVALUE:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "min_value";
 				value[i - 1] = Int64GetDatumFast(new.min_value);
 				break;
 			case SEQ_COL_CACHE:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "cache_value";
 				value[i - 1] = Int64GetDatumFast(new.cache_value);
 				break;
 			case SEQ_COL_LOG:
-				typnam->name = "int8";
+				typnam->typeid = INT8OID;
 				coldef->colname = "log_cnt";
 				value[i - 1] = Int64GetDatum((int64) 1);
 				break;
 			case SEQ_COL_CYCLE:
-				typnam->name = "bool";
+				typnam->typeid = BOOLOID;
 				coldef->colname = "is_cycled";
 				value[i - 1] = BoolGetDatum(new.is_cycled);
 				break;
 			case SEQ_COL_CALLED:
-				typnam->name = "bool";
+				typnam->typeid = BOOLOID;
 				coldef->colname = "is_called";
 				value[i - 1] = BoolGetDatum(false);
 				break;
@@ -170,24 +177,21 @@ DefineSequence(CreateSeqStmt *seq)
 		stmt->tableElts = lappend(stmt->tableElts, coldef);
 	}
 
-	stmt->relname = seq->seqname;
-	stmt->inhRelnames = NIL;
+	stmt->relation = seq->sequence;
+	stmt->inhRelations = NIL;
 	stmt->constraints = NIL;
-	stmt->istemp = seq->istemp;
-	stmt->hasoids = false;
+	stmt->hasoids = MUST_NOT_HAVE_OIDS;
+	stmt->oncommit = ONCOMMIT_NOOP;
+	stmt->tablespacename = NULL;
 
-	DefineRelation(stmt, RELKIND_SEQUENCE);
+	seqoid = DefineRelation(stmt, RELKIND_SEQUENCE);
 
-	rel = heap_openr(seq->seqname, AccessExclusiveLock);
+	rel = heap_open(seqoid, AccessExclusiveLock);
 	tupDesc = RelationGetDescr(rel);
 
 	/* Initialize first page of relation with special magic number */
 
 	buf = ReadBuffer(rel, P_NEW);
-
-	if (!BufferIsValid(buf))
-		elog(ERROR, "DefineSequence: ReadBuffer failed");
-
 	Assert(BufferGetBlockNumber(buf) == 0);
 
 	page = (PageHeader) BufferGetPage(buf);
@@ -201,7 +205,7 @@ DefineSequence(CreateSeqStmt *seq)
 
 	/* Now form & insert sequence tuple */
 	tuple = heap_formtuple(tupDesc, value, null);
-	heap_insert(rel, tuple);
+	simple_heap_insert(rel, tuple);
 
 	Assert(ItemPointerGetOffsetNumber(&(tuple->t_self)) == FirstOffsetNumber);
 
@@ -209,27 +213,29 @@ DefineSequence(CreateSeqStmt *seq)
 	 * Two special hacks here:
 	 *
 	 * 1. Since VACUUM does not process sequences, we have to force the tuple
-	 * to have xmin = FrozenTransactionId now.  Otherwise it would become
-	 * invisible to SELECTs after 2G transactions.  It is okay to do this
+	 * to have xmin = FrozenTransactionId now.	Otherwise it would become
+	 * invisible to SELECTs after 2G transactions.	It is okay to do this
 	 * because if the current transaction aborts, no other xact will ever
 	 * examine the sequence tuple anyway.
 	 *
 	 * 2. Even though heap_insert emitted a WAL log record, we have to emit
 	 * an XLOG_SEQ_LOG record too, since (a) the heap_insert record will
 	 * not have the right xmin, and (b) REDO of the heap_insert record
-	 * would re-init page and sequence magic number would be lost.  This
+	 * would re-init page and sequence magic number would be lost.	This
 	 * means two log records instead of one :-(
 	 */
 	LockBuffer(buf, BUFFER_LOCK_EXCLUSIVE);
+
 	START_CRIT_SECTION();
 
 	{
 		/*
-		 * Note that the "tuple" structure is still just a local tuple record
-		 * created by heap_formtuple; its t_data pointer doesn't point at the
-		 * disk buffer.  To scribble on the disk buffer we need to fetch the
-		 * item pointer.  But do the same to the local tuple, since that will
-		 * be the source for the WAL log record, below.
+		 * Note that the "tuple" structure is still just a local tuple
+		 * record created by heap_formtuple; its t_data pointer doesn't
+		 * point at the disk buffer.  To scribble on the disk buffer we
+		 * need to fetch the item pointer.	But do the same to the local
+		 * tuple, since that will be the source for the WAL log record,
+		 * below.
 		 */
 		ItemId		itemId;
 		Item		item;
@@ -237,13 +243,15 @@ DefineSequence(CreateSeqStmt *seq)
 		itemId = PageGetItemId((Page) page, FirstOffsetNumber);
 		item = PageGetItem((Page) page, itemId);
 
-		((HeapTupleHeader) item)->t_xmin = FrozenTransactionId;
+		HeapTupleHeaderSetXmin((HeapTupleHeader) item, FrozenTransactionId);
 		((HeapTupleHeader) item)->t_infomask |= HEAP_XMIN_COMMITTED;
 
-		tuple->t_data->t_xmin = FrozenTransactionId;
+		HeapTupleHeaderSetXmin(tuple->t_data, FrozenTransactionId);
 		tuple->t_data->t_infomask |= HEAP_XMIN_COMMITTED;
 	}
 
+	/* XLOG stuff */
+	if (!rel->rd_istemp)
 	{
 		xl_seq_rec	xlrec;
 		XLogRecPtr	recptr;
@@ -269,8 +277,9 @@ DefineSequence(CreateSeqStmt *seq)
 		recptr = XLogInsert(RM_SEQ_ID, XLOG_SEQ_LOG | XLOG_NO_TRAN, rdata);
 
 		PageSetLSN(page, recptr);
-		PageSetSUI(page, ThisStartUpID);
+		PageSetTLI(page, ThisTimeLineID);
 	}
+
 	END_CRIT_SECTION();
 
 	LockBuffer(buf, BUFFER_LOCK_UNLOCK);
@@ -278,13 +287,91 @@ DefineSequence(CreateSeqStmt *seq)
 	heap_close(rel, NoLock);
 }
 
+/*
+ * AlterSequence
+ *
+ * Modify the definition of a sequence relation
+ */
+void
+AlterSequence(AlterSeqStmt *stmt)
+{
+	SeqTable	elm;
+	Relation	seqrel;
+	Buffer		buf;
+	Page		page;
+	Form_pg_sequence seq;
+	FormData_pg_sequence new;
+
+	/* open and AccessShareLock sequence */
+	init_sequence(stmt->sequence, &elm, &seqrel);
+
+	/* allow ALTER to sequence owner only */
+	if (!pg_class_ownercheck(elm->relid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   stmt->sequence->relname);
+
+	/* lock page' buffer and read tuple into new sequence structure */
+	seq = read_info(elm, seqrel, &buf);
+	page = BufferGetPage(buf);
+
+	/* Copy old values of options into workspace */
+	memcpy(&new, seq, sizeof(FormData_pg_sequence));
+
+	/* Check and set new values */
+	init_params(stmt->options, &new, false);
+
+	/* Now okay to update the on-disk tuple */
+	memcpy(seq, &new, sizeof(FormData_pg_sequence));
+
+	/* Clear local cache so that we don't think we have cached numbers */
+	elm->last = new.last_value; /* last returned number */
+	elm->cached = new.last_value;		/* last cached number (forget
+										 * cached values) */
+
+	START_CRIT_SECTION();
+
+	/* XLOG stuff */
+	if (!seqrel->rd_istemp)
+	{
+		xl_seq_rec	xlrec;
+		XLogRecPtr	recptr;
+		XLogRecData rdata[2];
+
+		xlrec.node = seqrel->rd_node;
+		rdata[0].buffer = InvalidBuffer;
+		rdata[0].data = (char *) &xlrec;
+		rdata[0].len = sizeof(xl_seq_rec);
+		rdata[0].next = &(rdata[1]);
+
+		rdata[1].buffer = InvalidBuffer;
+		rdata[1].data = (char *) page + ((PageHeader) page)->pd_upper;
+		rdata[1].len = ((PageHeader) page)->pd_special -
+			((PageHeader) page)->pd_upper;
+		rdata[1].next = NULL;
+
+		recptr = XLogInsert(RM_SEQ_ID, XLOG_SEQ_LOG | XLOG_NO_TRAN, rdata);
+
+		PageSetLSN(page, recptr);
+		PageSetTLI(page, ThisTimeLineID);
+	}
+
+	END_CRIT_SECTION();
+
+	LockBuffer(buf, BUFFER_LOCK_UNLOCK);
+
+	WriteBuffer(buf);
+
+	relation_close(seqrel, NoLock);
+}
+
 
 Datum
 nextval(PG_FUNCTION_ARGS)
 {
 	text	   *seqin = PG_GETARG_TEXT_P(0);
-	char	   *seqname = get_seq_name(seqin);
+	RangeVar   *sequence;
 	SeqTable	elm;
+	Relation	seqrel;
 	Buffer		buf;
 	Page		page;
 	Form_pg_sequence seq;
@@ -300,23 +387,27 @@ nextval(PG_FUNCTION_ARGS)
 				rescnt = 0;
 	bool		logit = false;
 
-	if (pg_aclcheck(seqname, GetUserId(), ACL_UPDATE) != ACLCHECK_OK)
-		elog(ERROR, "%s.nextval: you don't have permissions to set sequence %s",
-			 seqname, seqname);
+	sequence = makeRangeVarFromNameList(textToQualifiedNameList(seqin,
+															 "nextval"));
 
 	/* open and AccessShareLock sequence */
-	elm = init_sequence("nextval", seqname);
+	init_sequence(sequence, &elm, &seqrel);
 
-	pfree(seqname);
+	if (pg_class_aclcheck(elm->relid, GetUserId(), ACL_UPDATE) != ACLCHECK_OK)
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("permission denied for sequence %s",
+						sequence->relname)));
 
 	if (elm->last != elm->cached)		/* some numbers were cached */
 	{
 		elm->last += elm->increment;
+		relation_close(seqrel, NoLock);
 		PG_RETURN_INT64(elm->last);
 	}
 
-	seq = read_info("nextval", elm, &buf);		/* lock page' buffer and
-												 * read tuple */
+	/* lock page' buffer and read tuple */
+	seq = read_info(elm, seqrel, &buf);
 	page = BufferGetPage(buf);
 
 	last = next = result = seq->last_value;
@@ -334,14 +425,14 @@ nextval(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Decide whether we should emit a WAL log record.  If so, force up
+	 * Decide whether we should emit a WAL log record.	If so, force up
 	 * the fetch count to grab SEQ_LOG_VALS more values than we actually
 	 * need to cache.  (These will then be usable without logging.)
 	 *
-	 * If this is the first nextval after a checkpoint, we must force
-	 * a new WAL record to be written anyway, else replay starting from the
+	 * If this is the first nextval after a checkpoint, we must force a new
+	 * WAL record to be written anyway, else replay starting from the
 	 * checkpoint would fail to advance the sequence past the logged
-	 * values.  In this case we may as well fetch extra values.
+	 * values.	In this case we may as well fetch extra values.
 	 */
 	if (log < fetch)
 	{
@@ -376,8 +467,15 @@ nextval(PG_FUNCTION_ARGS)
 				if (rescnt > 0)
 					break;		/* stop fetching */
 				if (!seq->is_cycled)
-					elog(ERROR, "%s.nextval: reached MAXVALUE (" INT64_FORMAT ")",
-						 elm->name, maxv);
+				{
+					char		buf[100];
+
+					snprintf(buf, sizeof(buf), INT64_FORMAT, maxv);
+					ereport(ERROR,
+					  (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+					   errmsg("nextval: reached maximum value of sequence \"%s\" (%s)",
+							  sequence->relname, buf)));
+				}
 				next = minv;
 			}
 			else
@@ -392,8 +490,15 @@ nextval(PG_FUNCTION_ARGS)
 				if (rescnt > 0)
 					break;		/* stop fetching */
 				if (!seq->is_cycled)
-					elog(ERROR, "%s.nextval: reached MINVALUE (" INT64_FORMAT ")",
-						 elm->name, minv);
+				{
+					char		buf[100];
+
+					snprintf(buf, sizeof(buf), INT64_FORMAT, minv);
+					ereport(ERROR,
+					  (errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+					   errmsg("nextval: reached minimum value of sequence \"%s\" (%s)",
+							  sequence->relname, buf)));
+				}
 				next = maxv;
 			}
 			else
@@ -418,21 +523,25 @@ nextval(PG_FUNCTION_ARGS)
 	elm->cached = last;			/* last fetched number */
 
 	START_CRIT_SECTION();
-	if (logit)
+
+	/* XLOG stuff */
+	if (logit && !seqrel->rd_istemp)
 	{
 		xl_seq_rec	xlrec;
 		XLogRecPtr	recptr;
 		XLogRecData rdata[2];
 
-		xlrec.node = elm->rel->rd_node;
+		xlrec.node = seqrel->rd_node;
 		rdata[0].buffer = InvalidBuffer;
 		rdata[0].data = (char *) &xlrec;
 		rdata[0].len = sizeof(xl_seq_rec);
 		rdata[0].next = &(rdata[1]);
 
+		/* set values that will be saved in xlog */
 		seq->last_value = next;
 		seq->is_called = true;
 		seq->log_cnt = 0;
+
 		rdata[1].buffer = InvalidBuffer;
 		rdata[1].data = (char *) page + ((PageHeader) page)->pd_upper;
 		rdata[1].len = ((PageHeader) page)->pd_special -
@@ -442,19 +551,21 @@ nextval(PG_FUNCTION_ARGS)
 		recptr = XLogInsert(RM_SEQ_ID, XLOG_SEQ_LOG | XLOG_NO_TRAN, rdata);
 
 		PageSetLSN(page, recptr);
-		PageSetSUI(page, ThisStartUpID);
+		PageSetTLI(page, ThisTimeLineID);
 	}
 
 	/* update on-disk data */
 	seq->last_value = last;		/* last fetched number */
 	seq->is_called = true;
 	seq->log_cnt = log;			/* how much is logged */
+
 	END_CRIT_SECTION();
 
 	LockBuffer(buf, BUFFER_LOCK_UNLOCK);
 
-	if (WriteBuffer(buf) == STATUS_ERROR)
-		elog(ERROR, "%s.nextval: WriteBuffer failed", elm->name);
+	WriteBuffer(buf);
+
+	relation_close(seqrel, NoLock);
 
 	PG_RETURN_INT64(result);
 }
@@ -463,24 +574,32 @@ Datum
 currval(PG_FUNCTION_ARGS)
 {
 	text	   *seqin = PG_GETARG_TEXT_P(0);
-	char	   *seqname = get_seq_name(seqin);
+	RangeVar   *sequence;
 	SeqTable	elm;
+	Relation	seqrel;
 	int64		result;
 
-	if (pg_aclcheck(seqname, GetUserId(), ACL_SELECT) != ACLCHECK_OK)
-		elog(ERROR, "%s.currval: you don't have permissions to read sequence %s",
-			 seqname, seqname);
+	sequence = makeRangeVarFromNameList(textToQualifiedNameList(seqin,
+															 "currval"));
 
 	/* open and AccessShareLock sequence */
-	elm = init_sequence("currval", seqname);
+	init_sequence(sequence, &elm, &seqrel);
+
+	if (pg_class_aclcheck(elm->relid, GetUserId(), ACL_SELECT) != ACLCHECK_OK)
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("permission denied for sequence %s",
+						sequence->relname)));
 
 	if (elm->increment == 0)	/* nextval/read_info were not called */
-		elog(ERROR, "%s.currval is not yet defined in this session",
-			 seqname);
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("currval of sequence \"%s\" is not yet defined in this session",
+						sequence->relname)));
 
 	result = elm->last;
 
-	pfree(seqname);
+	relation_close(seqrel, NoLock);
 
 	PG_RETURN_INT64(result);
 }
@@ -499,24 +618,39 @@ currval(PG_FUNCTION_ARGS)
  * sequence.
  */
 static void
-do_setval(char *seqname, int64 next, bool iscalled)
+do_setval(RangeVar *sequence, int64 next, bool iscalled)
 {
 	SeqTable	elm;
+	Relation	seqrel;
 	Buffer		buf;
 	Form_pg_sequence seq;
 
-	if (pg_aclcheck(seqname, GetUserId(), ACL_UPDATE) != ACLCHECK_OK)
-		elog(ERROR, "%s.setval: you don't have permissions to set sequence %s",
-			 seqname, seqname);
-
 	/* open and AccessShareLock sequence */
-	elm = init_sequence("setval", seqname);
-	seq = read_info("setval", elm, &buf);		/* lock page' buffer and
-												 * read tuple */
+	init_sequence(sequence, &elm, &seqrel);
+
+	if (pg_class_aclcheck(elm->relid, GetUserId(), ACL_UPDATE) != ACLCHECK_OK)
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("permission denied for sequence %s",
+						sequence->relname)));
+
+	/* lock page' buffer and read tuple */
+	seq = read_info(elm, seqrel, &buf);
 
 	if ((next < seq->min_value) || (next > seq->max_value))
-		elog(ERROR, "%s.setval: value " INT64_FORMAT " is out of bounds (" INT64_FORMAT "," INT64_FORMAT ")",
-			 seqname, next, seq->min_value, seq->max_value);
+	{
+		char		bufv[100],
+					bufm[100],
+					bufx[100];
+
+		snprintf(bufv, sizeof(bufv), INT64_FORMAT, next);
+		snprintf(bufm, sizeof(bufm), INT64_FORMAT, seq->min_value);
+		snprintf(bufx, sizeof(bufx), INT64_FORMAT, seq->max_value);
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("setval: value %s is out of bounds for sequence \"%s\" (%s..%s)",
+						bufv, sequence->relname, bufm, bufx)));
+	}
 
 	/* save info in local cache */
 	elm->last = next;			/* last returned number */
@@ -524,21 +658,26 @@ do_setval(char *seqname, int64 next, bool iscalled)
 								 * values) */
 
 	START_CRIT_SECTION();
+
+	/* XLOG stuff */
+	if (!seqrel->rd_istemp)
 	{
 		xl_seq_rec	xlrec;
 		XLogRecPtr	recptr;
 		XLogRecData rdata[2];
 		Page		page = BufferGetPage(buf);
 
-		xlrec.node = elm->rel->rd_node;
+		xlrec.node = seqrel->rd_node;
 		rdata[0].buffer = InvalidBuffer;
 		rdata[0].data = (char *) &xlrec;
 		rdata[0].len = sizeof(xl_seq_rec);
 		rdata[0].next = &(rdata[1]);
 
+		/* set values that will be saved in xlog */
 		seq->last_value = next;
 		seq->is_called = true;
 		seq->log_cnt = 0;
+
 		rdata[1].buffer = InvalidBuffer;
 		rdata[1].data = (char *) page + ((PageHeader) page)->pd_upper;
 		rdata[1].len = ((PageHeader) page)->pd_special -
@@ -548,20 +687,21 @@ do_setval(char *seqname, int64 next, bool iscalled)
 		recptr = XLogInsert(RM_SEQ_ID, XLOG_SEQ_LOG | XLOG_NO_TRAN, rdata);
 
 		PageSetLSN(page, recptr);
-		PageSetSUI(page, ThisStartUpID);
+		PageSetTLI(page, ThisTimeLineID);
 	}
+
 	/* save info in sequence relation */
 	seq->last_value = next;		/* last fetched number */
 	seq->is_called = iscalled;
 	seq->log_cnt = (iscalled) ? 0 : 1;
+
 	END_CRIT_SECTION();
 
 	LockBuffer(buf, BUFFER_LOCK_UNLOCK);
 
-	if (WriteBuffer(buf) == STATUS_ERROR)
-		elog(ERROR, "%s.setval: WriteBuffer failed", seqname);
+	WriteBuffer(buf);
 
-	pfree(seqname);
+	relation_close(seqrel, NoLock);
 }
 
 /*
@@ -573,9 +713,12 @@ setval(PG_FUNCTION_ARGS)
 {
 	text	   *seqin = PG_GETARG_TEXT_P(0);
 	int64		next = PG_GETARG_INT64(1);
-	char	   *seqname = get_seq_name(seqin);
+	RangeVar   *sequence;
 
-	do_setval(seqname, next, true);
+	sequence = makeRangeVarFromNameList(textToQualifiedNameList(seqin,
+															  "setval"));
+
+	do_setval(sequence, next, true);
 
 	PG_RETURN_INT64(next);
 }
@@ -590,69 +733,111 @@ setval_and_iscalled(PG_FUNCTION_ARGS)
 	text	   *seqin = PG_GETARG_TEXT_P(0);
 	int64		next = PG_GETARG_INT64(1);
 	bool		iscalled = PG_GETARG_BOOL(2);
-	char	   *seqname = get_seq_name(seqin);
+	RangeVar   *sequence;
 
-	do_setval(seqname, next, iscalled);
+	sequence = makeRangeVarFromNameList(textToQualifiedNameList(seqin,
+															  "setval"));
+
+	do_setval(sequence, next, iscalled);
 
 	PG_RETURN_INT64(next);
 }
 
+
 /*
- * Given a 'text' parameter to a sequence function, extract the actual
- * sequence name.  We downcase the name if it's not double-quoted,
- * and truncate it if it's too long.
- *
- * This is a kluge, really --- should be able to write nextval(seqrel).
+ * Given a relation name, open and lock the sequence.  p_elm and p_rel are
+ * output parameters.
  */
-static char *
-get_seq_name(text *seqin)
+static void
+init_sequence(RangeVar *relation, SeqTable *p_elm, Relation *p_rel)
 {
-	char	   *rawname = DatumGetCString(DirectFunctionCall1(textout,
-												PointerGetDatum(seqin)));
-	int			rawlen = strlen(rawname);
-	char	   *seqname;
+	Oid			relid = RangeVarGetRelid(relation, false);
+	TransactionId thisxid = GetTopTransactionId();
+	volatile SeqTable elm;
+	Relation	seqrel;
 
-	if (rawlen >= 2 &&
-		rawname[0] == '\"' && rawname[rawlen - 1] == '\"')
+	/*
+	 * Open the sequence relation.
+	 */
+	seqrel = relation_open(relid, NoLock);
+
+	if (seqrel->rd_rel->relkind != RELKIND_SEQUENCE)
+		ereport(ERROR,
+				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+				 errmsg("\"%s\" is not a sequence",
+						relation->relname)));
+
+	/* Look to see if we already have a seqtable entry for relation */
+	for (elm = seqtab; elm != NULL; elm = elm->next)
 	{
-		/* strip off quotes, keep case */
-		rawname[rawlen - 1] = '\0';
-		seqname = pstrdup(rawname + 1);
-		pfree(rawname);
+		if (elm->relid == relid)
+			break;
 	}
-	else
-	{
-		seqname = rawname;
 
+	/*
+	 * Allocate new seqtable entry if we didn't find one.
+	 *
+	 * NOTE: seqtable entries remain in the list for the life of a backend.
+	 * If the sequence itself is deleted then the entry becomes wasted
+	 * memory, but it's small enough that this should not matter.
+	 */
+	if (elm == NULL)
+	{
 		/*
-		 * It's important that this match the identifier downcasing code
-		 * used by backend/parser/scan.l.
+		 * Time to make a new seqtable entry.  These entries live as long
+		 * as the backend does, so we use plain malloc for them.
 		 */
-		for (; *rawname; rawname++)
-		{
-			if (isupper((unsigned char) *rawname))
-				*rawname = tolower((unsigned char) *rawname);
-		}
+		elm = (SeqTable) malloc(sizeof(SeqTableData));
+		if (elm == NULL)
+			ereport(ERROR,
+					(errcode(ERRCODE_OUT_OF_MEMORY),
+					 errmsg("out of memory")));
+		elm->relid = relid;
+		elm->xid = InvalidTransactionId;
+		/* increment is set to 0 until we do read_info (see currval) */
+		elm->last = elm->cached = elm->increment = 0;
+		elm->next = seqtab;
+		seqtab = elm;
 	}
 
-	/* Truncate name if it's overlength; again, should match scan.l */
-	if (strlen(seqname) >= NAMEDATALEN)
+	/*
+	 * If we haven't touched the sequence already in this transaction,
+	 * we need to acquire AccessShareLock.  We arrange for the lock to
+	 * be owned by the top transaction, so that we don't need to do it
+	 * more than once per xact.
+	 */
+	if (elm->xid != thisxid)
 	{
-#ifdef MULTIBYTE
-		int			len;
+		ResourceOwner currentOwner;
 
-		len = pg_mbcliplen(seqname, rawlen, NAMEDATALEN - 1);
-		seqname[len] = '\0';
-#else
-		seqname[NAMEDATALEN - 1] = '\0';
-#endif
+		currentOwner = CurrentResourceOwner;
+		PG_TRY();
+		{
+			CurrentResourceOwner = TopTransactionResourceOwner;
+
+			LockRelation(seqrel, AccessShareLock);
+		}
+		PG_CATCH();
+		{
+			/* Ensure CurrentResourceOwner is restored on error */
+			CurrentResourceOwner = currentOwner;
+			PG_RE_THROW();
+		}
+		PG_END_TRY();
+		CurrentResourceOwner = currentOwner;
+
+		/* Flag that we have a lock in the current xact. */
+		elm->xid = thisxid;
 	}
 
-	return seqname;
+	*p_elm = elm;
+	*p_rel = seqrel;
 }
 
+
+/* Given an opened relation, lock the page buffer and find the tuple */
 static Form_pg_sequence
-read_info(char *caller, SeqTable elm, Buffer *buf)
+read_info(SeqTable elm, Relation rel, Buffer *buf)
 {
 	PageHeader	page;
 	ItemId		lp;
@@ -660,21 +845,15 @@ read_info(char *caller, SeqTable elm, Buffer *buf)
 	sequence_magic *sm;
 	Form_pg_sequence seq;
 
-	if (elm->rel->rd_nblocks > 1)
-		elog(ERROR, "%s.%s: invalid number of blocks in sequence",
-			 elm->name, caller);
-
-	*buf = ReadBuffer(elm->rel, 0);
-	if (!BufferIsValid(*buf))
-		elog(ERROR, "%s.%s: ReadBuffer failed", elm->name, caller);
-
+	*buf = ReadBuffer(rel, 0);
 	LockBuffer(*buf, BUFFER_LOCK_EXCLUSIVE);
 
 	page = (PageHeader) BufferGetPage(*buf);
 	sm = (sequence_magic *) PageGetSpecialPointer(page);
 
 	if (sm->magic != SEQ_MAGIC)
-		elog(ERROR, "%s.%s: bad magic (%08X)", elm->name, caller, sm->magic);
+		elog(ERROR, "bad magic number in sequence \"%s\": %08X",
+			 RelationGetRelationName(rel), sm->magic);
 
 	lp = PageGetItemId(page, FirstOffsetNumber);
 	Assert(ItemIdIsUsed(lp));
@@ -687,206 +866,205 @@ read_info(char *caller, SeqTable elm, Buffer *buf)
 	return seq;
 }
 
-
-static SeqTable
-init_sequence(char *caller, char *name)
-{
-	SeqTable	elm,
-				prev = (SeqTable) NULL;
-	Relation	seqrel;
-
-	/* Look to see if we already have a seqtable entry for name */
-	for (elm = seqtab; elm != (SeqTable) NULL; elm = elm->next)
-	{
-		if (strcmp(elm->name, name) == 0)
-			break;
-		prev = elm;
-	}
-
-	/* If so, and if it's already been opened in this xact, just return it */
-	if (elm != (SeqTable) NULL && elm->rel != (Relation) NULL)
-		return elm;
-
-	/* Else open and check it */
-	seqrel = heap_openr(name, AccessShareLock);
-	if (seqrel->rd_rel->relkind != RELKIND_SEQUENCE)
-		elog(ERROR, "%s.%s: %s is not a sequence", name, caller, name);
-
-	if (elm != (SeqTable) NULL)
-	{
-		/*
-		 * We are using a seqtable entry left over from a previous xact;
-		 * must check for relid change.
-		 */
-		elm->rel = seqrel;
-		if (RelationGetRelid(seqrel) != elm->relid)
-		{
-			elog(NOTICE, "%s.%s: sequence was re-created",
-				 name, caller);
-			elm->relid = RelationGetRelid(seqrel);
-			elm->cached = elm->last = elm->increment = 0;
-		}
-	}
-	else
-	{
-		/*
-		 * Time to make a new seqtable entry.  These entries live as long
-		 * as the backend does, so we use plain malloc for them.
-		 */
-		elm = (SeqTable) malloc(sizeof(SeqTableData));
-		if (elm == NULL)
-			elog(ERROR, "Memory exhausted in init_sequence");
-		elm->name = strdup(name);
-		if (elm->name == NULL)
-			elog(ERROR, "Memory exhausted in init_sequence");
-		elm->rel = seqrel;
-		elm->relid = RelationGetRelid(seqrel);
-		elm->cached = elm->last = elm->increment = 0;
-		elm->next = (SeqTable) NULL;
-
-		if (seqtab == (SeqTable) NULL)
-			seqtab = elm;
-		else
-			prev->next = elm;
-	}
-
-	return elm;
-}
-
-
 /*
- * CloseSequences
- *				is called by xact mgr at commit/abort.
+ * init_params: process the options list of CREATE or ALTER SEQUENCE,
+ * and store the values into appropriate fields of *new.
+ *
+ * If isInit is true, fill any unspecified options with default values;
+ * otherwise, do not change existing options that aren't explicitly overridden.
  */
-void
-CloseSequences(void)
-{
-	SeqTable	elm;
-	Relation	rel;
-
-	for (elm = seqtab; elm != (SeqTable) NULL; elm = elm->next)
-	{
-		rel = elm->rel;
-		if (rel != (Relation) NULL)		/* opened in current xact */
-		{
-			elm->rel = (Relation) NULL;
-			heap_close(rel, AccessShareLock);
-		}
-	}
-}
-
-
 static void
-init_params(CreateSeqStmt *seq, Form_pg_sequence new)
+init_params(List *options, Form_pg_sequence new, bool isInit)
 {
 	DefElem    *last_value = NULL;
 	DefElem    *increment_by = NULL;
 	DefElem    *max_value = NULL;
 	DefElem    *min_value = NULL;
 	DefElem    *cache_value = NULL;
-	List	   *option;
+	DefElem    *is_cycled = NULL;
+	ListCell   *option;
 
-	new->is_cycled = false;
-	foreach(option, seq->options)
+	foreach(option, options)
 	{
 		DefElem    *defel = (DefElem *) lfirst(option);
 
 		if (strcmp(defel->defname, "increment") == 0)
+		{
+			if (increment_by)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
 			increment_by = defel;
-		else if (strcmp(defel->defname, "start") == 0)
+		}
+
+		/*
+		 * start is for a new sequence restart is for alter
+		 */
+		else if (strcmp(defel->defname, "start") == 0 ||
+				 strcmp(defel->defname, "restart") == 0)
+		{
+			if (last_value)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
 			last_value = defel;
+		}
 		else if (strcmp(defel->defname, "maxvalue") == 0)
+		{
+			if (max_value)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
 			max_value = defel;
+		}
 		else if (strcmp(defel->defname, "minvalue") == 0)
+		{
+			if (min_value)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
 			min_value = defel;
+		}
 		else if (strcmp(defel->defname, "cache") == 0)
+		{
+			if (cache_value)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
 			cache_value = defel;
+		}
 		else if (strcmp(defel->defname, "cycle") == 0)
 		{
-			if (defel->arg != (Node *) NULL)
-				elog(ERROR, "DefineSequence: CYCLE ??");
-			new->is_cycled = true;
+			if (is_cycled)
+				ereport(ERROR,
+						(errcode(ERRCODE_SYNTAX_ERROR),
+						 errmsg("conflicting or redundant options")));
+			is_cycled = defel;
 		}
 		else
-			elog(ERROR, "DefineSequence: option \"%s\" not recognized",
+			elog(ERROR, "option \"%s\" not recognized",
 				 defel->defname);
 	}
 
-	if (increment_by == (DefElem *) NULL)		/* INCREMENT BY */
+	/* INCREMENT BY */
+	if (increment_by != NULL)
+	{
+		new->increment_by = defGetInt64(increment_by);
+		if (new->increment_by == 0)
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("INCREMENT must not be zero")));
+	}
+	else if (isInit)
 		new->increment_by = 1;
-	else if ((new->increment_by = get_param(increment_by)) == 0)
-		elog(ERROR, "DefineSequence: can't INCREMENT by 0");
 
-	if (max_value == (DefElem *) NULL)	/* MAXVALUE */
+	/* CYCLE */
+	if (is_cycled != NULL)
+	{
+		new->is_cycled = intVal(is_cycled->arg);
+		Assert(new->is_cycled == false || new->is_cycled == true);
+	}
+	else if (isInit)
+		new->is_cycled = false;
+
+	/* MAXVALUE (null arg means NO MAXVALUE) */
+	if (max_value != NULL && max_value->arg)
+		new->max_value = defGetInt64(max_value);
+	else if (isInit || max_value != NULL)
 	{
 		if (new->increment_by > 0)
 			new->max_value = SEQ_MAXVALUE;		/* ascending seq */
 		else
 			new->max_value = -1;	/* descending seq */
 	}
-	else
-		new->max_value = get_param(max_value);
 
-	if (min_value == (DefElem *) NULL)	/* MINVALUE */
+	/* MINVALUE (null arg means NO MINVALUE) */
+	if (min_value != NULL && min_value->arg)
+		new->min_value = defGetInt64(min_value);
+	else if (isInit || min_value != NULL)
 	{
 		if (new->increment_by > 0)
 			new->min_value = 1; /* ascending seq */
 		else
 			new->min_value = SEQ_MINVALUE;		/* descending seq */
 	}
-	else
-		new->min_value = get_param(min_value);
 
+	/* crosscheck min/max */
 	if (new->min_value >= new->max_value)
-		elog(ERROR, "DefineSequence: MINVALUE (" INT64_FORMAT ") can't be >= MAXVALUE (" INT64_FORMAT ")",
-			 new->min_value, new->max_value);
+	{
+		char		bufm[100],
+					bufx[100];
 
-	if (last_value == (DefElem *) NULL) /* START WITH */
+		snprintf(bufm, sizeof(bufm), INT64_FORMAT, new->min_value);
+		snprintf(bufx, sizeof(bufx), INT64_FORMAT, new->max_value);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("MINVALUE (%s) must be less than MAXVALUE (%s)",
+						bufm, bufx)));
+	}
+
+	/* START WITH */
+	if (last_value != NULL)
+	{
+		new->last_value = defGetInt64(last_value);
+		new->is_called = false;
+		new->log_cnt = 1;
+	}
+	else if (isInit)
 	{
 		if (new->increment_by > 0)
 			new->last_value = new->min_value;	/* ascending seq */
 		else
 			new->last_value = new->max_value;	/* descending seq */
+		new->is_called = false;
+		new->log_cnt = 1;
 	}
-	else
-		new->last_value = get_param(last_value);
 
+	/* crosscheck */
 	if (new->last_value < new->min_value)
-		elog(ERROR, "DefineSequence: START value (" INT64_FORMAT ") can't be < MINVALUE (" INT64_FORMAT ")",
-			 new->last_value, new->min_value);
+	{
+		char		bufs[100],
+					bufm[100];
+
+		snprintf(bufs, sizeof(bufs), INT64_FORMAT, new->last_value);
+		snprintf(bufm, sizeof(bufm), INT64_FORMAT, new->min_value);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			  errmsg("START value (%s) can't be less than MINVALUE (%s)",
+					 bufs, bufm)));
+	}
 	if (new->last_value > new->max_value)
-		elog(ERROR, "DefineSequence: START value (" INT64_FORMAT ") can't be > MAXVALUE (" INT64_FORMAT ")",
-			 new->last_value, new->max_value);
+	{
+		char		bufs[100],
+					bufm[100];
 
-	if (cache_value == (DefElem *) NULL)		/* CACHE */
+		snprintf(bufs, sizeof(bufs), INT64_FORMAT, new->last_value);
+		snprintf(bufm, sizeof(bufm), INT64_FORMAT, new->max_value);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+		   errmsg("START value (%s) can't be greater than MAXVALUE (%s)",
+				  bufs, bufm)));
+	}
+
+	/* CACHE */
+	if (cache_value != NULL)
+	{
+		new->cache_value = defGetInt64(cache_value);
+		if (new->cache_value <= 0)
+		{
+			char		buf[100];
+
+			snprintf(buf, sizeof(buf), INT64_FORMAT, new->cache_value);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					 errmsg("CACHE (%s) must be greater than zero",
+							buf)));
+		}
+	}
+	else if (isInit)
 		new->cache_value = 1;
-	else if ((new->cache_value = get_param(cache_value)) <= 0)
-		elog(ERROR, "DefineSequence: CACHE (" INT64_FORMAT ") can't be <= 0",
-			 new->cache_value);
-
 }
 
-static int64
-get_param(DefElem *def)
-{
-	if (def->arg == (Node *) NULL)
-		elog(ERROR, "DefineSequence: \"%s\" value unspecified", def->defname);
-
-	if (IsA(def->arg, Integer))
-		return (int64) intVal(def->arg);
-
-	/*
-	 * Values too large for int4 will be represented as Float constants by
-	 * the lexer.  Accept these if they are valid int8 strings.
-	 */
-	if (IsA(def->arg, Float))
-		return DatumGetInt64(DirectFunctionCall1(int8in,
-									 CStringGetDatum(strVal(def->arg))));
-
-	/* Shouldn't get here unless parser messed up */
-	elog(ERROR, "DefineSequence: \"%s\" value must be integer", def->defname);
-	return 0;					/* not reached; keep compiler quiet */
-}
 
 void
 seq_redo(XLogRecPtr lsn, XLogRecord *record)
@@ -901,7 +1079,7 @@ seq_redo(XLogRecPtr lsn, XLogRecord *record)
 	sequence_magic *sm;
 
 	if (info != XLOG_SEQ_LOG)
-		elog(STOP, "seq_redo: unknown op code %u", info);
+		elog(PANIC, "seq_redo: unknown op code %u", info);
 
 	reln = XLogOpenRelation(true, RM_SEQ_ID, xlrec->node);
 	if (!RelationIsValid(reln))
@@ -909,8 +1087,8 @@ seq_redo(XLogRecPtr lsn, XLogRecord *record)
 
 	buffer = XLogReadBuffer(true, reln, 0);
 	if (!BufferIsValid(buffer))
-		elog(STOP, "seq_redo: can't read block of %u/%u",
-			 xlrec->node.tblNode, xlrec->node.relNode);
+		elog(PANIC, "seq_redo: can't read block 0 of rel %u/%u/%u",
+		   xlrec->node.spcNode, xlrec->node.dbNode, xlrec->node.relNode);
 
 	page = (Page) BufferGetPage(buffer);
 
@@ -925,11 +1103,12 @@ seq_redo(XLogRecPtr lsn, XLogRecord *record)
 	itemsz = MAXALIGN(itemsz);
 	if (PageAddItem(page, (Item) item, itemsz,
 					FirstOffsetNumber, LP_USED) == InvalidOffsetNumber)
-		elog(STOP, "seq_redo: failed to add item to page");
+		elog(PANIC, "seq_redo: failed to add item to page");
 
 	PageSetLSN(page, lsn);
-	PageSetSUI(page, ThisStartUpID);
-	UnlockAndWriteBuffer(buffer);
+	PageSetTLI(page, ThisTimeLineID);
+	LockBuffer(buffer, BUFFER_LOCK_UNLOCK);
+	WriteBuffer(buffer);
 }
 
 void
@@ -951,6 +1130,6 @@ seq_desc(char *buf, uint8 xl_info, char *rec)
 		return;
 	}
 
-	sprintf(buf + strlen(buf), "node %u/%u",
-			xlrec->node.tblNode, xlrec->node.relNode);
+	sprintf(buf + strlen(buf), "rel %u/%u/%u",
+			xlrec->node.spcNode, xlrec->node.dbNode, xlrec->node.relNode);
 }
