@@ -3,12 +3,12 @@
  * tlist.c
  *	  Target list manipulation routines
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/optimizer/util/tlist.c,v 1.52 2002/06/20 20:29:31 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/optimizer/util/tlist.c,v 1.68 2004/12/31 22:00:23 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -17,10 +17,11 @@
 #include "nodes/makefuncs.h"
 #include "optimizer/tlist.h"
 #include "optimizer/var.h"
+#include "parser/parse_expr.h"
 
 
 /*****************************************************************************
- *	---------- RELATION node target list routines ----------
+ *		Target list creation and searching utilities
  *****************************************************************************/
 
 /*
@@ -31,7 +32,7 @@
 TargetEntry *
 tlistentry_member(Node *node, List *targetlist)
 {
-	List	   *temp;
+	ListCell   *temp;
 
 	foreach(temp, targetlist)
 	{
@@ -58,7 +59,7 @@ matching_tlist_expr(Node *node, List *targetlist)
 	if (tlentry)
 		return tlentry->expr;
 
-	return (Node *) NULL;
+	return NULL;
 }
 #endif
 
@@ -76,73 +77,44 @@ tlist_member(Node *node, List *targetlist)
 	if (tlentry)
 		return tlentry->resdom;
 
-	return (Resdom *) NULL;
-}
-
-/*
- * add_var_to_tlist
- *	  Creates a targetlist entry corresponding to the supplied var node
- *	  'var' and adds the new targetlist entry to the targetlist field of
- *	  'rel'.  No entry is created if 'var' is already in the tlist.
- */
-void
-add_var_to_tlist(RelOptInfo *rel, Var *var)
-{
-	if (!tlistentry_member((Node *) var, rel->targetlist))
-	{
-		/* XXX is copyObject necessary here? */
-		rel->targetlist = lappend(rel->targetlist,
-							   create_tl_element((Var *) copyObject(var),
-										   length(rel->targetlist) + 1));
-	}
+	return NULL;
 }
 
 /*
  * create_tl_element
  *	  Creates a target list entry node and its associated (resdom var) pair
  *	  with its resdom number equal to 'resdomno'.
+ *
+ * Note: the argument is almost always a Var, but occasionally not.
  */
 TargetEntry *
 create_tl_element(Var *var, int resdomno)
 {
+	Oid			vartype;
+	int32		vartypmod;
+
+	if (IsA(var, Var))
+	{
+		vartype = var->vartype;
+		vartypmod = var->vartypmod;
+	}
+	else
+	{
+		vartype = exprType((Node *) var);
+		vartypmod = exprTypmod((Node *) var);
+	}
 	return makeTargetEntry(makeResdom(resdomno,
-									  var->vartype,
-									  var->vartypmod,
+									  vartype,
+									  vartypmod,
 									  NULL,
 									  false),
-						   (Node *) var);
+						   (Expr *) var);
 }
 
 /*****************************************************************************
  *		---------- GENERAL target list routines ----------
  *****************************************************************************/
 
-/*
- * new_unsorted_tlist
- *	  Creates a copy of a target list by creating new resdom nodes
- *	  without sort information.
- *
- * 'targetlist' is the target list to be copied.
- *
- * Returns the resulting target list.
- *
- */
-List *
-new_unsorted_tlist(List *targetlist)
-{
-	List	   *new_targetlist = (List *) copyObject((Node *) targetlist);
-	List	   *x;
-
-	foreach(x, new_targetlist)
-	{
-		TargetEntry *tle = (TargetEntry *) lfirst(x);
-
-		tle->resdom->reskey = 0;
-		tle->resdom->reskeyop = (Oid) 0;
-	}
-	return new_targetlist;
-}
-
 /*
  * flatten_tlist
  *	  Create a target list that only contains unique variables.
@@ -165,7 +137,7 @@ flatten_tlist(List *tlist)
 	List	   *new_tlist;
 
 	new_tlist = add_to_flat_tlist(NIL, vlist);
-	freeList(vlist);
+	list_free(vlist);
 	return new_tlist;
 }
 
@@ -181,12 +153,12 @@ flatten_tlist(List *tlist)
 List *
 add_to_flat_tlist(List *tlist, List *vars)
 {
-	int			next_resdomno = length(tlist) + 1;
-	List	   *v;
+	int			next_resdomno = list_length(tlist) + 1;
+	ListCell   *v;
 
 	foreach(v, vars)
 	{
-		Var		   *var = lfirst(v);
+		Var		   *var = (Var *) lfirst(v);
 
 		if (!tlistentry_member((Node *) var, tlist))
 		{
@@ -204,15 +176,6 @@ add_to_flat_tlist(List *tlist, List *vars)
 	return tlist;
 }
 
-Var *
-get_expr(TargetEntry *tle)
-{
-	Assert(tle != NULL);
-	Assert(tle->expr != NULL);
-
-	return (Var *) tle->expr;
-}
-
 /*
  * get_sortgroupclause_tle
  *		Find the targetlist entry matching the given SortClause
@@ -226,7 +189,7 @@ get_sortgroupclause_tle(SortClause *sortClause,
 						List *targetList)
 {
 	Index		refnumber = sortClause->tleSortGroupRef;
-	List	   *l;
+	ListCell   *l;
 
 	foreach(l, targetList)
 	{
@@ -236,7 +199,7 @@ get_sortgroupclause_tle(SortClause *sortClause,
 			return tle;
 	}
 
-	elog(ERROR, "get_sortgroupclause_tle: ORDER/GROUP BY expression not found in targetlist");
+	elog(ERROR, "ORDER/GROUP BY expression not found in targetlist");
 	return NULL;				/* keep compiler quiet */
 }
 
@@ -253,5 +216,27 @@ get_sortgroupclause_expr(SortClause *sortClause, List *targetList)
 {
 	TargetEntry *tle = get_sortgroupclause_tle(sortClause, targetList);
 
-	return tle->expr;
+	return (Node *) tle->expr;
+}
+
+/*
+ * get_sortgrouplist_exprs
+ *		Given a list of SortClauses (or GroupClauses), build a list
+ *		of the referenced targetlist expressions.
+ */
+List *
+get_sortgrouplist_exprs(List *sortClauses, List *targetList)
+{
+	List	   *result = NIL;
+	ListCell   *l;
+
+	foreach(l, sortClauses)
+	{
+		SortClause *sortcl = (SortClause *) lfirst(l);
+		Node	   *sortexpr;
+
+		sortexpr = get_sortgroupclause_expr(sortcl, targetList);
+		result = lappend(result, sortexpr);
+	}
+	return result;
 }
