@@ -3,12 +3,12 @@
  * date.c
  *	  implements DATE and TIME data types specified in SQL-92 standard
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994-5, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/date.c,v 1.73.2.4 2003/02/13 17:04:24 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/adt/date.c,v 1.104.4.1 2005/04/23 22:53:26 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -17,11 +17,12 @@
 
 #include <ctype.h>
 #include <limits.h>
-#include <time.h>
 #include <float.h>
 
 #include "access/hash.h"
+#include "libpq/pqformat.h"
 #include "miscadmin.h"
+#include "parser/scansup.h"
 #include "utils/builtins.h"
 #include "utils/date.h"
 #include "utils/nabstime.h"
@@ -36,10 +37,10 @@
 #endif
 
 
-static int	time2tm(TimeADT time, struct tm * tm, fsec_t *fsec);
-static int	timetz2tm(TimeTzADT *time, struct tm * tm, fsec_t *fsec, int *tzp);
-static int	tm2time(struct tm * tm, fsec_t fsec, TimeADT *result);
-static int	tm2timetz(struct tm * tm, fsec_t fsec, int tz, TimeTzADT *result);
+static int	time2tm(TimeADT time, struct pg_tm * tm, fsec_t *fsec);
+static int	timetz2tm(TimeTzADT *time, struct pg_tm * tm, fsec_t *fsec, int *tzp);
+static int	tm2time(struct pg_tm * tm, fsec_t fsec, TimeADT *result);
+static int	tm2timetz(struct pg_tm * tm, fsec_t fsec, int tz, TimeTzADT *result);
 static void AdjustTimeForTypmod(TimeADT *time, int32 typmod);
 
 /*****************************************************************************
@@ -54,23 +55,26 @@ Datum
 date_in(PG_FUNCTION_ARGS)
 {
 	char	   *str = PG_GETARG_CSTRING(0);
-	DateADT		date;
+	DateADT date;
 	fsec_t		fsec;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	int			tzp;
 	int			dtype;
 	int			nf;
+	int			dterr;
 	char	   *field[MAXDATEFIELDS];
 	int			ftype[MAXDATEFIELDS];
 	char		lowstr[MAXDATELEN + 1];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad date external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	 || (DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tzp) != 0))
-		elog(ERROR, "Bad date external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tzp);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "date");
 
 	switch (dtype)
 	{
@@ -78,8 +82,10 @@ date_in(PG_FUNCTION_ARGS)
 			break;
 
 		case DTK_CURRENT:
-			elog(ERROR, "Date CURRENT no longer supported"
-				 "\n\tdate_in() internal coding error");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("date/time value \"current\" is no longer supported")));
+
 			GetCurrentDateTime(tm);
 			break;
 
@@ -88,10 +94,11 @@ date_in(PG_FUNCTION_ARGS)
 			break;
 
 		default:
-			elog(ERROR, "Unrecognized date external representation '%s'", str);
+			DateTimeParseError(DTERR_BAD_FORMAT, str, "date");
+			break;
 	}
 
-	date = (date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1));
+	date = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(date);
 }
@@ -102,13 +109,13 @@ date_in(PG_FUNCTION_ARGS)
 Datum
 date_out(PG_FUNCTION_ARGS)
 {
-	DateADT		date = PG_GETARG_DATEADT(0);
+	DateADT date = PG_GETARG_DATEADT(0);
 	char	   *result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	char		buf[MAXDATELEN + 1];
 
-	j2date((date + date2j(2000, 1, 1)),
+	j2date(date +POSTGRES_EPOCH_JDATE,
 		   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
 
 	EncodeDateOnly(tm, DateStyle, buf);
@@ -117,6 +124,36 @@ date_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		date_recv			- converts external binary format to date
+ */
+Datum
+date_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+	PG_RETURN_DATEADT((DateADT) pq_getmsgint(buf, sizeof(DateADT)));
+}
+
+/*
+ *		date_send			- converts date to binary format
+ */
+Datum
+date_send(PG_FUNCTION_ARGS)
+{
+	DateADT date = PG_GETARG_DATEADT(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, date, sizeof(date));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
+/*
+ * Comparison functions for dates
+ */
+
 Datum
 date_eq(PG_FUNCTION_ARGS)
 {
@@ -236,59 +273,428 @@ date_mii(PG_FUNCTION_ARGS)
 	PG_RETURN_DATEADT(dateVal - days);
 }
 
-#if NOT_USED
-/* date_pl_interval() and date_mi_interval() are probably
- * better implmented by converting the input date
- * to timestamp without time zone. So that is what we do
- * in pg_proc.h - thomas 2002-03-11
+/*
+ * Internal routines for promoting date to timestamp and timestamp with
+ * time zone
  */
 
+#ifdef HAVE_INT64_TIMESTAMP
+/* date is days since 2000, timestamp is microseconds since same... */
+#define date2timestamp(dateVal) \
+	((Timestamp) ((dateVal) * INT64CONST(86400000000)))
+#else
+/* date is days since 2000, timestamp is seconds since same... */
+#define date2timestamp(dateVal) \
+	((Timestamp) ((dateVal) * 86400.0))
+#endif
+
+static TimestampTz
+date2timestamptz(DateADT dateVal)
+{
+	TimestampTz result;
+	struct pg_tm tt,
+			   *tm = &tt;
+	int			tz;
+
+	j2date(dateVal + POSTGRES_EPOCH_JDATE,
+		   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
+
+	tm->tm_hour = 0;
+	tm->tm_min = 0;
+	tm->tm_sec = 0;
+	tz = DetermineLocalTimeZone(tm);
+
+#ifdef HAVE_INT64_TIMESTAMP
+	result = (dateVal * INT64CONST(86400000000))
+		+ (tz * INT64CONST(1000000));
+#else
+	result = dateVal * 86400.0 + tz;
+#endif
+
+	return result;
+}
+
+
+/*
+ * Crosstype comparison functions for dates
+ */
+
+Datum
+date_eq_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) == 0);
+}
+
+Datum
+date_ne_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) != 0);
+}
+
+Datum
+date_lt_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) < 0);
+}
+
+Datum
+date_gt_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) > 0);
+}
+
+Datum
+date_le_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) <= 0);
+}
+
+Datum
+date_ge_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) >= 0);
+}
+
+Datum
+date_cmp_timestamp(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	Timestamp	dt2 = PG_GETARG_TIMESTAMP(1);
+	Timestamp	dt1;
+
+	dt1 = date2timestamp(dateVal);
+
+	PG_RETURN_INT32(timestamp_cmp_internal(dt1, dt2));
+}
+
+Datum
+date_eq_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) == 0);
+}
+
+Datum
+date_ne_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) != 0);
+}
+
+Datum
+date_lt_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) < 0);
+}
+
+Datum
+date_gt_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) > 0);
+}
+
+Datum
+date_le_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) <= 0);
+}
+
+Datum
+date_ge_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) >= 0);
+}
+
+Datum
+date_cmp_timestamptz(PG_FUNCTION_ARGS)
+{
+	DateADT		dateVal = PG_GETARG_DATEADT(0);
+	TimestampTz dt2 = PG_GETARG_TIMESTAMPTZ(1);
+	TimestampTz dt1;
+
+	dt1 = date2timestamptz(dateVal);
+
+	PG_RETURN_INT32(timestamptz_cmp_internal(dt1, dt2));
+}
+
+Datum
+timestamp_eq_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) == 0);
+}
+
+Datum
+timestamp_ne_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) != 0);
+}
+
+Datum
+timestamp_lt_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) < 0);
+}
+
+Datum
+timestamp_gt_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) > 0);
+}
+
+Datum
+timestamp_le_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) <= 0);
+}
+
+Datum
+timestamp_ge_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_BOOL(timestamp_cmp_internal(dt1, dt2) >= 0);
+}
+
+Datum
+timestamp_cmp_date(PG_FUNCTION_ARGS)
+{
+	Timestamp	dt1 = PG_GETARG_TIMESTAMP(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	Timestamp	dt2;
+
+	dt2 = date2timestamp(dateVal);
+
+	PG_RETURN_INT32(timestamp_cmp_internal(dt1, dt2));
+}
+
+Datum
+timestamptz_eq_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) == 0);
+}
+
+Datum
+timestamptz_ne_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) != 0);
+}
+
+Datum
+timestamptz_lt_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) < 0);
+}
+
+Datum
+timestamptz_gt_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) > 0);
+}
+
+Datum
+timestamptz_le_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) <= 0);
+}
+
+Datum
+timestamptz_ge_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_BOOL(timestamptz_cmp_internal(dt1, dt2) >= 0);
+}
+
+Datum
+timestamptz_cmp_date(PG_FUNCTION_ARGS)
+{
+	TimestampTz dt1 = PG_GETARG_TIMESTAMPTZ(0);
+	DateADT		dateVal = PG_GETARG_DATEADT(1);
+	TimestampTz dt2;
+
+	dt2 = date2timestamptz(dateVal);
+
+	PG_RETURN_INT32(timestamptz_cmp_internal(dt1, dt2));
+}
+
+
 /* Add an interval to a date, giving a new date.
  * Must handle both positive and negative intervals.
+ *
+ * We implement this by promoting the date to timestamp (without time zone)
+ * and then using the timestamp plus interval function.
  */
 Datum
 date_pl_interval(PG_FUNCTION_ARGS)
 {
 	DateADT		dateVal = PG_GETARG_DATEADT(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
-	struct tm	tt,
-			   *tm = &tt;
+	Timestamp	dateStamp;
 
-	if (span->month != 0)
-	{
-		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
-		tm->tm_mon += span->month;
-		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
-	}
-	if (span->time != 0)
-		dateVal += (span->time / 86400e0);
+	dateStamp = date2timestamp(dateVal);
 
-	PG_RETURN_DATEADT(dateVal);
+	return DirectFunctionCall2(timestamp_pl_interval,
+							   TimestampGetDatum(dateStamp),
+							   PointerGetDatum(span));
 }
 
 /* Subtract an interval from a date, giving a new date.
  * Must handle both positive and negative intervals.
+ *
+ * We implement this by promoting the date to timestamp (without time zone)
+ * and then using the timestamp minus interval function.
  */
 Datum
 date_mi_interval(PG_FUNCTION_ARGS)
 {
 	DateADT		dateVal = PG_GETARG_DATEADT(0);
 	Interval   *span = PG_GETARG_INTERVAL_P(1);
-	struct tm	tt,
-			   *tm = &tt;
+	Timestamp	dateStamp;
 
-	if (span->month != 0)
-	{
-		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
-		tm->tm_mon -= span->month;
-		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
-	}
-	if (span->time != 0)
-		dateVal -= (span->time / 86400e0);
+	dateStamp = date2timestamp(dateVal);
 
-	PG_RETURN_DATEADT(dateVal);
+	return DirectFunctionCall2(timestamp_mi_interval,
+							   TimestampGetDatum(dateStamp),
+							   PointerGetDatum(span));
 }
-#endif
 
 /* date_timestamp()
  * Convert date to timestamp data type.
@@ -299,13 +705,7 @@ date_timestamp(PG_FUNCTION_ARGS)
 	DateADT		dateVal = PG_GETARG_DATEADT(0);
 	Timestamp	result;
 
-#ifdef HAVE_INT64_TIMESTAMP
-	/* date is days since 2000, timestamp is microseconds since same... */
-	result = dateVal * INT64CONST(86400000000);
-#else
-	/* date is days since 2000, timestamp is seconds since same... */
-	result = dateVal * 86400.0;
-#endif
+	result = date2timestamp(dateVal);
 
 	PG_RETURN_TIMESTAMP(result);
 }
@@ -317,19 +717,21 @@ date_timestamp(PG_FUNCTION_ARGS)
 Datum
 timestamp_date(PG_FUNCTION_ARGS)
 {
-	Timestamp	timestamp = PG_GETARG_TIMESTAMP(0);
+	Timestamp timestamp = PG_GETARG_TIMESTAMP(0);
 	DateADT		result;
+	struct pg_tm tt,
+			   *tm = &tt;
+	fsec_t		fsec;
 
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-#ifdef HAVE_INT64_TIMESTAMP
-	/* Microseconds to days */
-	result = (timestamp / INT64CONST(86400000000));
-#else
-	/* Seconds to days */
-	result = (timestamp / 86400.0);
-#endif
+	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) !=0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
+
+	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(result);
 }
@@ -343,37 +745,8 @@ date_timestamptz(PG_FUNCTION_ARGS)
 {
 	DateADT		dateVal = PG_GETARG_DATEADT(0);
 	TimestampTz result;
-	struct tm	tt,
-			   *tm = &tt;
 
-	j2date((dateVal + date2j(2000, 1, 1)),
-		   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
-
-	if (IS_VALID_UTIME(tm->tm_year, tm->tm_mon, tm->tm_mday))
-	{
-		int			tz;
-
-		tm->tm_hour = 0;
-		tm->tm_min = 0;
-		tm->tm_sec = 0;
-		tz = DetermineLocalTimeZone(tm);
-
-#ifdef HAVE_INT64_TIMESTAMP
-		result = (dateVal * INT64CONST(86400000000))
-			+ (tz * INT64CONST(1000000));
-#else
-		result = dateVal * 86400.0 + tz;
-#endif
-	}
-	else
-	{
-		/* Outside of range for timezone support, so assume UTC */
-#ifdef HAVE_INT64_TIMESTAMP
-		result = (dateVal * INT64CONST(86400000000));
-#else
-		result = dateVal * 86400.0;
-#endif
-	}
+	result = date2timestamptz(dateVal);
 
 	PG_RETURN_TIMESTAMP(result);
 }
@@ -387,7 +760,7 @@ timestamptz_date(PG_FUNCTION_ARGS)
 {
 	TimestampTz timestamp = PG_GETARG_TIMESTAMP(0);
 	DateADT		result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	fsec_t		fsec;
 	int			tz;
@@ -396,10 +769,12 @@ timestamptz_date(PG_FUNCTION_ARGS)
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
+	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) !=0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
-	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(result);
 }
@@ -413,7 +788,7 @@ abstime_date(PG_FUNCTION_ARGS)
 {
 	AbsoluteTime abstime = PG_GETARG_ABSOLUTETIME(0);
 	DateADT		result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	int			tz;
 
@@ -422,7 +797,9 @@ abstime_date(PG_FUNCTION_ARGS)
 		case INVALID_ABSTIME:
 		case NOSTART_ABSTIME:
 		case NOEND_ABSTIME:
-			elog(ERROR, "Unable to convert reserved abstime value to date");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			   errmsg("cannot convert reserved abstime value to date")));
 
 			/*
 			 * pretend to drop through to make compiler think that result
@@ -431,7 +808,7 @@ abstime_date(PG_FUNCTION_ARGS)
 
 		default:
 			abstime2tm(abstime, &tz, tm, NULL);
-			result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+			result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 			break;
 	}
 
@@ -446,7 +823,7 @@ Datum
 date_text(PG_FUNCTION_ARGS)
 {
 	/* Input is a Date, but may as well leave it in Datum form */
-	Datum		date = PG_GETARG_DATUM(0);
+	Datum date = PG_GETARG_DATUM(0);
 	text	   *result;
 	char	   *str;
 	int			len;
@@ -481,7 +858,10 @@ text_date(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad date external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type date: \"%s\"",
+						VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -509,20 +889,24 @@ time_in(PG_FUNCTION_ARGS)
 	int32		typmod = PG_GETARG_INT32(2);
 	TimeADT		result;
 	fsec_t		fsec;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
+	int			tz;
 	int			nf;
+	int			dterr;
 	char		lowstr[MAXDATELEN + 1];
 	char	   *field[MAXDATEFIELDS];
 	int			dtype;
 	int			ftype[MAXDATEFIELDS];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad time external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	 || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, NULL) != 0))
-		elog(ERROR, "Bad time external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "time");
 
 	tm2time(tm, fsec, &result);
 	AdjustTimeForTypmod(&result, typmod);
@@ -534,7 +918,7 @@ time_in(PG_FUNCTION_ARGS)
  * Convert a tm structure to a time data type.
  */
 static int
-tm2time(struct tm * tm, fsec_t fsec, TimeADT *result)
+tm2time(struct pg_tm * tm, fsec_t fsec, TimeADT *result)
 {
 #ifdef HAVE_INT64_TIMESTAMP
 	*result = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
@@ -551,7 +935,7 @@ tm2time(struct tm * tm, fsec_t fsec, TimeADT *result)
  *	local time zone. If out of this range, leave as GMT. - tgl 97/05/27
  */
 static int
-time2tm(TimeADT time, struct tm * tm, fsec_t *fsec)
+time2tm(TimeADT time, struct pg_tm * tm, fsec_t *fsec)
 {
 #ifdef HAVE_INT64_TIMESTAMP
 	tm->tm_hour = (time / INT64CONST(3600000000));
@@ -579,7 +963,7 @@ time_out(PG_FUNCTION_ARGS)
 {
 	TimeADT		time = PG_GETARG_TIMEADT(0);
 	char	   *result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	fsec_t		fsec;
 	char		buf[MAXDATELEN + 1];
@@ -591,6 +975,43 @@ time_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		time_recv			- converts external binary format to time
+ *
+ * We make no attempt to provide compatibility between int and float
+ * time representations ...
+ */
+Datum
+time_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+#ifdef HAVE_INT64_TIMESTAMP
+	PG_RETURN_TIMEADT((TimeADT) pq_getmsgint64(buf));
+#else
+	PG_RETURN_TIMEADT((TimeADT) pq_getmsgfloat8(buf));
+#endif
+}
+
+/*
+ *		time_send			- converts time to binary format
+ */
+Datum
+time_send(PG_FUNCTION_ARGS)
+{
+	TimeADT		time = PG_GETARG_TIMEADT(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+#ifdef HAVE_INT64_TIMESTAMP
+	pq_sendint64(&buf, time);
+#else
+	pq_sendfloat8(&buf, time);
+#endif
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
 /* time_scale()
  * Adjust time type for specified scale factor.
  * Used by PostgreSQL type system to stuff columns.
@@ -642,17 +1063,17 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 #else
 	/* note MAX_TIME_PRECISION differs in this case */
 	static const double TimeScales[MAX_TIME_PRECISION + 1] = {
-		1,
-		10,
-		100,
-		1000,
-		10000,
-		100000,
-		1000000,
-		10000000,
-		100000000,
-		1000000000,
-		10000000000
+		1.0,
+		10.0,
+		100.0,
+		1000.0,
+		10000.0,
+		100000.0,
+		1000000.0,
+		10000000.0,
+		100000000.0,
+		1000000000.0,
+		10000000000.0
 	};
 #endif
 
@@ -661,9 +1082,9 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 		/*
 		 * Note: this round-to-nearest code is not completely consistent
 		 * about rounding values that are exactly halfway between integral
-		 * values.  On most platforms, rint() will implement round-to-nearest-even,
-		 * but the integer code always rounds up (away from zero).  Is it
-		 * worth trying to be consistent?
+		 * values.	On most platforms, rint() will implement
+		 * round-to-nearest-even, but the integer code always rounds up
+		 * (away from zero).  Is it worth trying to be consistent?
 		 */
 #ifdef HAVE_INT64_TIMESTAMP
 		if (*time >= INT64CONST(0))
@@ -673,8 +1094,8 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 		}
 		else
 		{
-			*time = - ((((- *time) + TimeOffsets[typmod]) / TimeScales[typmod])
-					   * TimeScales[typmod]);
+			*time = -((((-*time) + TimeOffsets[typmod]) / TimeScales[typmod])
+					  * TimeScales[typmod]);
 		}
 #else
 		*time = (rint(((double) *time) * TimeScales[typmod])
@@ -901,17 +1322,19 @@ overlaps_time(PG_FUNCTION_ARGS)
 Datum
 timestamp_time(PG_FUNCTION_ARGS)
 {
-	Timestamp	timestamp = PG_GETARG_TIMESTAMP(0);
+	Timestamp timestamp = PG_GETARG_TIMESTAMP(0);
 	TimeADT		result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	fsec_t		fsec;
 
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) != 0)
-		elog(ERROR, "Unable to convert timestamp to time");
+	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) !=0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 #ifdef HAVE_INT64_TIMESTAMP
 
@@ -936,7 +1359,7 @@ timestamptz_time(PG_FUNCTION_ARGS)
 {
 	TimestampTz timestamp = PG_GETARG_TIMESTAMP(0);
 	TimeADT		result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	int			tz;
 	fsec_t		fsec;
@@ -945,8 +1368,10 @@ timestamptz_time(PG_FUNCTION_ARGS)
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamptz to time");
+	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) !=0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 #ifdef HAVE_INT64_TIMESTAMP
 
@@ -969,7 +1394,7 @@ timestamptz_time(PG_FUNCTION_ARGS)
 Datum
 datetime_timestamp(PG_FUNCTION_ARGS)
 {
-	DateADT		date = PG_GETARG_DATEADT(0);
+	DateADT date = PG_GETARG_DATEADT(0);
 	TimeADT		time = PG_GETARG_TIMEADT(1);
 	Timestamp	result;
 
@@ -999,6 +1424,11 @@ time_interval(PG_FUNCTION_ARGS)
 
 /* interval_time()
  * Convert interval to time data type.
+ *
+ * This is defined as producing the fractional-day portion of the interval.
+ * Therefore, we can just ignore the months field.	It is not real clear
+ * what to do with negative intervals, but we choose to subtract the floor,
+ * so that, say, '-2 hours' becomes '22:00:00'.
  */
 Datum
 interval_time(PG_FUNCTION_ARGS)
@@ -1007,15 +1437,23 @@ interval_time(PG_FUNCTION_ARGS)
 	TimeADT		result;
 
 #ifdef HAVE_INT64_TIMESTAMP
-	result = span->time;
-	if ((result >= INT64CONST(86400000000))
-		|| (result <= INT64CONST(-86400000000)))
-		result -= (result / INT64CONST(1000000) * INT64CONST(1000000));
-#else
-	Interval	span1;
+	int64		days;
 
 	result = span->time;
-	TMODULO(result, span1.time, 86400e0);
+	if (result >= INT64CONST(86400000000))
+	{
+		days = result / INT64CONST(86400000000);
+		result -= days * INT64CONST(86400000000);
+	}
+	else if (result < 0)
+	{
+		days = (-result + INT64CONST(86400000000) - 1) / INT64CONST(86400000000);
+		result += days * INT64CONST(86400000000);
+	}
+#else
+	result = span->time;
+	if (result >= 86400e0 || result < 0)
+		result -= floor(result / 86400e0) * 86400e0;
 #endif
 
 	PG_RETURN_TIMEADT(result);
@@ -1093,18 +1531,6 @@ time_mi_interval(PG_FUNCTION_ARGS)
 	PG_RETURN_TIMEADT(result);
 }
 
-/* interval_pl_time()
- * Add time to interval.
- */
-Datum
-interval_pl_time(PG_FUNCTION_ARGS)
-{
-	Datum		span = PG_GETARG_DATUM(0);
-	Datum		time = PG_GETARG_DATUM(1);
-
-	return DirectFunctionCall2(time_pl_interval, time, span);
-}
-
 
 /* time_text()
  * Convert time to text data type.
@@ -1148,7 +1574,10 @@ text_time(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad time external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type time: \"%s\"",
+						VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -1173,20 +1602,11 @@ time_part(PG_FUNCTION_ARGS)
 	float8		result;
 	int			type,
 				val;
-	int			i;
-	char	   *up,
-			   *lp,
-				lowunits[MAXDATELEN + 1];
+	char	   *lowunits;
 
-	if (VARSIZE(units) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "TIME units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
-	up = VARDATA(units);
-	lp = lowunits;
-	for (i = 0; i < (VARSIZE(units) - VARHDRSZ); i++)
-		*lp++ = tolower((unsigned char) *up++);
-	*lp = '\0';
+	lowunits = downcase_truncate_identifier(VARDATA(units),
+											VARSIZE(units) - VARHDRSZ,
+											false);
 
 	type = DecodeUnits(0, lowunits, &val);
 	if (type == UNKNOWN_FIELD)
@@ -1195,7 +1615,7 @@ time_part(PG_FUNCTION_ARGS)
 	if (type == UNITS)
 	{
 		fsec_t		fsec;
-		struct tm	tt,
+		struct pg_tm tt,
 				   *tm = &tt;
 
 		time2tm(time, tm, &fsec);
@@ -1246,9 +1666,12 @@ time_part(PG_FUNCTION_ARGS)
 			case DTK_CENTURY:
 			case DTK_MILLENNIUM:
 			default:
-				elog(ERROR, "TIME units '%s' not supported",
-					 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("\"time\" units \"%s\" not recognized",
+							 DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 				result = 0;
 		}
 	}
@@ -1262,9 +1685,11 @@ time_part(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		elog(ERROR, "TIME units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"time\" units \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
 		result = 0;
 	}
 
@@ -1280,7 +1705,7 @@ time_part(PG_FUNCTION_ARGS)
  * Convert a tm structure to a time data type.
  */
 static int
-tm2timetz(struct tm * tm, fsec_t fsec, int tz, TimeTzADT *result)
+tm2timetz(struct pg_tm * tm, fsec_t fsec, int tz, TimeTzADT *result)
 {
 #ifdef HAVE_INT64_TIMESTAMP
 	result->time = ((((((tm->tm_hour * 60) + tm->tm_min) * 60) + tm->tm_sec)
@@ -1304,22 +1729,24 @@ timetz_in(PG_FUNCTION_ARGS)
 	int32		typmod = PG_GETARG_INT32(2);
 	TimeTzADT  *result;
 	fsec_t		fsec;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	int			tz;
 	int			nf;
+	int			dterr;
 	char		lowstr[MAXDATELEN + 1];
 	char	   *field[MAXDATEFIELDS];
 	int			dtype;
 	int			ftype[MAXDATEFIELDS];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad time with time zone"
-			 " external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	  || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz) != 0))
-		elog(ERROR, "Bad time external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "time with time zone");
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 	tm2timetz(tm, fsec, tz, result);
@@ -1333,7 +1760,7 @@ timetz_out(PG_FUNCTION_ARGS)
 {
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(0);
 	char	   *result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	fsec_t		fsec;
 	int			tz;
@@ -1346,11 +1773,52 @@ timetz_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		timetz_recv			- converts external binary format to timetz
+ */
+Datum
+timetz_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+	TimeTzADT  *time;
+
+	time = (TimeTzADT *) palloc(sizeof(TimeTzADT));
+
+#ifdef HAVE_INT64_TIMESTAMP
+	time->time = pq_getmsgint64(buf);
+#else
+	time->time = pq_getmsgfloat8(buf);
+#endif
+	time->zone = pq_getmsgint(buf, sizeof(time->zone));
+
+	PG_RETURN_TIMETZADT_P(time);
+}
+
+/*
+ *		timetz_send			- converts timetz to binary format
+ */
+Datum
+timetz_send(PG_FUNCTION_ARGS)
+{
+	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+#ifdef HAVE_INT64_TIMESTAMP
+	pq_sendint64(&buf, time->time);
+#else
+	pq_sendfloat8(&buf, time->time);
+#endif
+	pq_sendint(&buf, time->zone, sizeof(time->zone));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
 /* timetz2tm()
  * Convert TIME WITH TIME ZONE data type to POSIX time structure.
  */
 static int
-timetz2tm(TimeTzADT *time, struct tm * tm, fsec_t *fsec, int *tzp)
+timetz2tm(TimeTzADT *time, struct pg_tm * tm, fsec_t *fsec, int *tzp)
 {
 #ifdef HAVE_INT64_TIMESTAMP
 	int64		trem = time->time;
@@ -1401,12 +1869,20 @@ timetz_scale(PG_FUNCTION_ARGS)
 static int
 timetz_cmp_internal(TimeTzADT *time1, TimeTzADT *time2)
 {
+	/* Primary sort is by true (GMT-equivalent) time */
+#ifdef HAVE_INT64_TIMESTAMP
+	int64		t1,
+				t2;
+
+	t1 = time1->time + (time1->zone * INT64CONST(1000000));
+	t2 = time2->time + (time2->zone * INT64CONST(1000000));
+#else
 	double		t1,
 				t2;
 
-	/* Primary sort is by true (GMT-equivalent) time */
 	t1 = time1->time + time1->zone;
 	t2 = time2->time + time2->zone;
+#endif
 
 	if (t1 > t2)
 		return 1;
@@ -1509,12 +1985,13 @@ timetz_larger(PG_FUNCTION_ARGS)
 {
 	TimeTzADT  *time1 = PG_GETARG_TIMETZADT_P(0);
 	TimeTzADT  *time2 = PG_GETARG_TIMETZADT_P(1);
+	TimeTzADT  *result;
 
-	if (DatumGetBool(DirectFunctionCall2(timetz_gt,
-										 TimeTzADTPGetDatum(time1),
-										 TimeTzADTPGetDatum(time2))))
-		PG_RETURN_TIMETZADT_P(time1);
-	PG_RETURN_TIMETZADT_P(time2);
+	if (timetz_cmp_internal(time1, time2) > 0)
+		result = time1;
+	else
+		result = time2;
+	PG_RETURN_TIMETZADT_P(result);
 }
 
 Datum
@@ -1522,12 +1999,13 @@ timetz_smaller(PG_FUNCTION_ARGS)
 {
 	TimeTzADT  *time1 = PG_GETARG_TIMETZADT_P(0);
 	TimeTzADT  *time2 = PG_GETARG_TIMETZADT_P(1);
+	TimeTzADT  *result;
 
-	if (DatumGetBool(DirectFunctionCall2(timetz_lt,
-										 TimeTzADTPGetDatum(time1),
-										 TimeTzADTPGetDatum(time2))))
-		PG_RETURN_TIMETZADT_P(time1);
-	PG_RETURN_TIMETZADT_P(time2);
+	if (timetz_cmp_internal(time1, time2) < 0)
+		result = time1;
+	else
+		result = time2;
+	PG_RETURN_TIMETZADT_P(result);
 }
 
 /* timetz_pl_interval()
@@ -1741,7 +2219,7 @@ time_timetz(PG_FUNCTION_ARGS)
 {
 	TimeADT		time = PG_GETARG_TIMEADT(0);
 	TimeTzADT  *result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	fsec_t		fsec;
 	int			tz;
@@ -1767,7 +2245,7 @@ timestamptz_timetz(PG_FUNCTION_ARGS)
 {
 	TimestampTz timestamp = PG_GETARG_TIMESTAMP(0);
 	TimeTzADT  *result;
-	struct tm	tt,
+	struct pg_tm tt,
 			   *tm = &tt;
 	int			tz;
 	fsec_t		fsec;
@@ -1776,8 +2254,10 @@ timestamptz_timetz(PG_FUNCTION_ARGS)
 	if (TIMESTAMP_NOT_FINITE(timestamp))
 		PG_RETURN_NULL();
 
-	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamptz to timetz");
+	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) !=0)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
@@ -1796,15 +2276,15 @@ timestamptz_timetz(PG_FUNCTION_ARGS)
 Datum
 datetimetz_timestamptz(PG_FUNCTION_ARGS)
 {
-	DateADT		date = PG_GETARG_DATEADT(0);
+	DateADT date = PG_GETARG_DATEADT(0);
 	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(1);
 	TimestampTz result;
 
 #ifdef HAVE_INT64_TIMESTAMP
-	result = (((date * INT64CONST(86400000000)) + time->time)
+	result = (((date *INT64CONST(86400000000)) +time->time)
 			  + (time->zone * INT64CONST(1000000)));
 #else
-	result = (((date * 86400.0) + time->time) + time->zone);
+	result = (((date *86400.0) +time->time) + time->zone);
 #endif
 
 	PG_RETURN_TIMESTAMP(result);
@@ -1853,7 +2333,10 @@ text_timetz(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad timetz external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type time with time zone: \"%s\"",
+						VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -1878,20 +2361,11 @@ timetz_part(PG_FUNCTION_ARGS)
 	float8		result;
 	int			type,
 				val;
-	int			i;
-	char	   *up,
-			   *lp,
-				lowunits[MAXDATELEN + 1];
+	char	   *lowunits;
 
-	if (VARSIZE(units) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "TIMETZ units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
-	up = VARDATA(units);
-	lp = lowunits;
-	for (i = 0; i < (VARSIZE(units) - VARHDRSZ); i++)
-		*lp++ = tolower((unsigned char) *up++);
-	*lp = '\0';
+	lowunits = downcase_truncate_identifier(VARDATA(units),
+											VARSIZE(units) - VARHDRSZ,
+											false);
 
 	type = DecodeUnits(0, lowunits, &val);
 	if (type == UNKNOWN_FIELD)
@@ -1902,7 +2376,7 @@ timetz_part(PG_FUNCTION_ARGS)
 		double		dummy;
 		int			tz;
 		fsec_t		fsec;
-		struct tm	tt,
+		struct pg_tm tt,
 				   *tm = &tt;
 
 		timetz2tm(time, tm, &fsec, &tz);
@@ -1910,17 +2384,18 @@ timetz_part(PG_FUNCTION_ARGS)
 		switch (val)
 		{
 			case DTK_TZ:
-				result = tz;
+				result = -tz;
 				break;
 
 			case DTK_TZ_MINUTE:
-				result = tz / 60;
-				TMODULO(result, dummy, 60e0);
+				result = -tz;
+				result /= 60;
+				FMODULO(result, dummy, 60e0);
 				break;
 
 			case DTK_TZ_HOUR:
-				dummy = tz;
-				TMODULO(dummy, result, 3600e0);
+				dummy = -tz;
+				FMODULO(dummy, result, 3600e0);
 				break;
 
 			case DTK_MICROSEC:
@@ -1964,25 +2439,31 @@ timetz_part(PG_FUNCTION_ARGS)
 			case DTK_CENTURY:
 			case DTK_MILLENNIUM:
 			default:
-				elog(ERROR, "TIMETZ units '%s' not supported",
-					 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("\"time with time zone\" units \"%s\" not recognized",
+							 DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 				result = 0;
 		}
 	}
 	else if ((type == RESERV) && (val == DTK_EPOCH))
 	{
 #ifdef HAVE_INT64_TIMESTAMP
-		result = ((time->time / 1000000e0) - time->zone);
+		result = ((time->time / 1000000e0) + time->zone);
 #else
-		result = (time->time - time->zone);
+		result = (time->time + time->zone);
 #endif
 	}
 	else
 	{
-		elog(ERROR, "TIMETZ units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			errmsg("\"time with time zone\" units \"%s\" not recognized",
+				   DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 		result = 0;
 	}
 
@@ -2001,20 +2482,11 @@ timetz_zone(PG_FUNCTION_ARGS)
 	int			tz;
 	int			type,
 				val;
-	int			i;
-	char	   *up,
-			   *lp,
-				lowzone[MAXDATELEN + 1];
+	char	   *lowzone;
 
-	if (VARSIZE(zone) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Time zone '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-												 PointerGetDatum(zone))));
-	up = VARDATA(zone);
-	lp = lowzone;
-	for (i = 0; i < (VARSIZE(zone) - VARHDRSZ); i++)
-		*lp++ = tolower((unsigned char) *up++);
-	*lp = '\0';
+	lowzone = downcase_truncate_identifier(VARDATA(zone),
+										   VARSIZE(zone) - VARHDRSZ,
+										   false);
 
 	type = DecodeSpecial(0, lowzone, &val);
 
@@ -2041,7 +2513,10 @@ timetz_zone(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		elog(ERROR, "Time zone '%s' not recognized", lowzone);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("time zone \"%s\" not recognized", lowzone)));
+
 		PG_RETURN_NULL();
 	}
 
@@ -2060,9 +2535,11 @@ timetz_izone(PG_FUNCTION_ARGS)
 	int			tz;
 
 	if (zone->month != 0)
-		elog(ERROR, "INTERVAL time zone '%s' not legal (month specified)",
-			 DatumGetCString(DirectFunctionCall1(interval_out,
-												 PointerGetDatum(zone))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"interval\" time zone \"%s\" not valid",
+						DatumGetCString(DirectFunctionCall1(interval_out,
+											  PointerGetDatum(zone))))));
 
 #ifdef HAVE_INT64_TIMESTAMP
 	tz = -(zone->time / INT64CONST(1000000));
