@@ -1,4 +1,4 @@
-/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/execute.c,v 1.26.2.6 2006/04/24 09:46:32 meskes Exp $ */
+/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/execute.c,v 1.38.4.1 2005/03/18 10:01:14 meskes Exp $ */
 
 /*
  * The aim is to get a simpler inteface to the database routines.
@@ -66,10 +66,68 @@ quote_postgres(char *arg, int lineno)
 	res[ri++] = '\'';
 	res[ri] = '\0';
 
-	ECPGfree(arg);
 	return res;
 }
 
+#if defined(__GNUC__) && (defined (__powerpc__) || defined(__AMD64__))
+#define APREF ap
+#else
+#define APREF *ap
+#endif
+
+void
+ECPGget_variable(va_list APREF, enum ECPGttype type, struct variable * var, bool indicator)
+{
+	var->type = type;
+	var->pointer = va_arg(APREF, char *);
+
+	var->varcharsize = va_arg(APREF, long);
+	var->arrsize = va_arg(APREF, long);
+	var->offset = va_arg(APREF, long);
+
+	if (var->arrsize == 0 || var->varcharsize == 0)
+		var->value = *((char **) (var->pointer));
+	else
+		var->value = var->pointer;
+
+	/*
+	 * negative values are used to indicate an array without given bounds
+	 */
+	/* reset to zero for us */
+	if (var->arrsize < 0)
+		var->arrsize = 0;
+	if (var->varcharsize < 0)
+		var->varcharsize = 0;
+
+	var->next = NULL;
+
+	if (indicator)
+	{
+		var->ind_type = va_arg(APREF, enum ECPGttype);
+		var->ind_pointer = va_arg(APREF, char *);
+		var->ind_varcharsize = va_arg(APREF, long);
+		var->ind_arrsize = va_arg(APREF, long);
+		var->ind_offset = va_arg(APREF, long);
+
+		if (var->ind_type != ECPGt_NO_INDICATOR
+			&& (var->ind_arrsize == 0 || var->ind_varcharsize == 0))
+			var->ind_value = *((char **) (var->ind_pointer));
+		else
+			var->ind_value = var->ind_pointer;
+
+		/*
+		 * negative values are used to indicate an array without given
+		 * bounds
+		 */
+		/* reset to zero for us */
+		if (var->ind_arrsize < 0)
+			var->ind_arrsize = 0;
+		if (var->ind_varcharsize < 0)
+			var->ind_varcharsize = 0;
+	}
+}
+#undef APREF
+
 /*
  * create a list of variables
  * The variables are listed with input variables preceding outputvariables
@@ -119,8 +177,11 @@ create_statement(int lineno, int compat, int force_indicator, struct connection
 			if (!(var = (struct variable *) ECPGalloc(sizeof(struct variable), lineno)))
 				return false;
 
-			var->type = type;
-			var->pointer = va_arg(ap, char *);
+#if defined(__GNUC__) && (defined (__powerpc__) || defined(__AMD64__))
+			ECPGget_variable(ap, type, var, true);
+#else
+			ECPGget_variable(&ap, type, var, true);
+#endif
 
 			/* if variable is NULL, the statement hasn't been prepared */
 			if (var->pointer == NULL)
@@ -130,48 +191,6 @@ create_statement(int lineno, int compat, int force_indicator, struct connection
 				return false;
 			}
 
-			var->varcharsize = va_arg(ap, long);
-			var->arrsize = va_arg(ap, long);
-			var->offset = va_arg(ap, long);
-
-			if (var->arrsize == 0 || var->varcharsize == 0)
-				var->value = *((char **) (var->pointer));
-			else
-				var->value = var->pointer;
-
-			/*
-			 * negative values are used to indicate an array without given
-			 * bounds
-			 */
-			/* reset to zero for us */
-			if (var->arrsize < 0)
-				var->arrsize = 0;
-			if (var->varcharsize < 0)
-				var->varcharsize = 0;
-
-			var->ind_type = va_arg(ap, enum ECPGttype);
-			var->ind_pointer = va_arg(ap, char *);
-			var->ind_varcharsize = va_arg(ap, long);
-			var->ind_arrsize = va_arg(ap, long);
-			var->ind_offset = va_arg(ap, long);
-			var->next = NULL;
-
-			if (var->ind_type != ECPGt_NO_INDICATOR
-				&& (var->ind_arrsize == 0 || var->ind_varcharsize == 0))
-				var->ind_value = *((char **) (var->ind_pointer));
-			else
-				var->ind_value = var->ind_pointer;
-
-			/*
-			 * negative values are used to indicate an array without given
-			 * bounds
-			 */
-			/* reset to zero for us */
-			if (var->ind_arrsize < 0)
-				var->ind_arrsize = 0;
-			if (var->ind_varcharsize < 0)
-				var->ind_varcharsize = 0;
-
 			for (ptr = *list; ptr && ptr->next; ptr = ptr->next);
 
 			if (ptr == NULL)
@@ -236,19 +255,19 @@ ECPGtypeinfocache_push(struct ECPGtype_information_cache ** cache, int oid, bool
 {
 	struct ECPGtype_information_cache *new_entry
 	= (struct ECPGtype_information_cache *) ECPGalloc(sizeof(struct ECPGtype_information_cache), lineno);
-	
+
 	new_entry->oid = oid;
 	new_entry->isarray = isarray;
 	new_entry->next = *cache;
 	*cache = new_entry;
 }
-		
+
 static enum ARRAY_TYPE
 ECPGis_type_an_array(int type, const struct statement * stmt, const struct variable * var)
 {
-	char	   	*array_query;
-	enum ARRAY_TYPE	isarray = ECPG_ARRAY_NOT_SET;
-	PGresult   	*query;
+	char	   *array_query;
+	enum ARRAY_TYPE isarray = ECPG_ARRAY_NOT_SET;
+	PGresult   *query;
 	struct ECPGtype_information_cache *cache_entry;
 
 	if ((stmt->connection->cache_head) == NULL)
@@ -310,14 +329,14 @@ ECPGis_type_an_array(int type, const struct statement * stmt, const struct varia
 		if (cache_entry->oid == type)
 			return cache_entry->isarray;
 	}
-	
+
 	array_query = (char *) ECPGalloc(strlen("select typlen from pg_type where oid= and typelem<>0") + 11, stmt->lineno);
 	sprintf(array_query, "select typlen from pg_type where oid=%d and typelem<>0", type);
 	query = PQexec(stmt->connection->connection, array_query);
 	ECPGfree(array_query);
-	if (PQresultStatus(query) == PGRES_TUPLES_OK )
+	if (PQresultStatus(query) == PGRES_TUPLES_OK)
 	{
-		if ( PQntuples(query) == 0 )
+		if (PQntuples(query) == 0)
 			isarray = ECPG_ARRAY_NONE;
 		else
 		{
@@ -334,7 +353,7 @@ ECPGis_type_an_array(int type, const struct statement * stmt, const struct varia
 	}
 	PQclear(query);
 	ECPGtypeinfocache_push(&(stmt->connection->cache_head), type, isarray, stmt->lineno);
-	ECPGlog("ECPGis_type_an_array line %d: TYPE database: %d C: %d array: %d\n", stmt->lineno, type, var->type, isarray);
+	ECPGlog("ECPGis_type_an_array line %d: TYPE database: %d C: %d array: %s\n", stmt->lineno, type, var->type, isarray ? "Yes" : "No");
 	return isarray;
 }
 
@@ -343,7 +362,7 @@ bool
 ECPGstore_result(const PGresult *results, int act_field,
 				 const struct statement * stmt, struct variable * var)
 {
-	enum ARRAY_TYPE		isarray;
+	enum ARRAY_TYPE isarray;
 	int			act_tuple,
 				ntuples = PQntuples(results);
 	bool		status = true;
@@ -359,7 +378,7 @@ ECPGstore_result(const PGresult *results, int act_field,
 		{
 			ECPGlog("ECPGstore_result line %d: Incorrect number of matches: %d don't fit into array of %d\n",
 					stmt->lineno, ntuples, var->arrsize);
-			ECPGraise(stmt->lineno, INFORMIX_MODE(stmt->compat)?ECPG_INFORMIX_SUBSELECT_NOT_ONE:ECPG_TOO_MANY_MATCHES, ECPG_SQLSTATE_CARDINALITY_VIOLATION, NULL);
+			ECPGraise(stmt->lineno, INFORMIX_MODE(stmt->compat) ? ECPG_INFORMIX_SUBSELECT_NOT_ONE : ECPG_TOO_MANY_MATCHES, ECPG_SQLSTATE_CARDINALITY_VIOLATION, NULL);
 			return false;
 		}
 	}
@@ -478,8 +497,8 @@ ECPGstore_result(const PGresult *results, int act_field,
 	return status;
 }
 
-static bool
-ECPGstore_input(const struct statement * stmt, const struct variable * var,
+bool
+ECPGstore_input(const int lineno, const bool force_indicator, const struct variable * var,
 				const char **tobeinserted_p, bool *malloced_p)
 {
 	char	   *mallocedval = NULL;
@@ -492,7 +511,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 #if 0
 	if (var->arrsize > 1 &&...)
 	{
-		ECPGraise(stmt->lineno, ECPG_ARRAY_INSERT, ECPG_SQLSTATE_DATATYPE_MISMATCH, NULL);
+		ECPGraise(lineno, ECPG_ARRAY_INSERT, ECPG_SQLSTATE_DATATYPE_MISMATCH, NULL);
 		return false;
 	}
 #endif
@@ -531,9 +550,9 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 			break;
 #endif   /* HAVE_LONG_LONG_INT_64 */
 		case ECPGt_NO_INDICATOR:
-			if (stmt->force_indicator == false)
+			if (force_indicator == false)
 			{
-				if (ECPGis_informix_null(var->type, var->value))
+				if (ECPGis_noind_null(var->type, var->value))
 					*tobeinserted_p = "null";
 			}
 			break;
@@ -547,7 +566,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				int			element;
 
 			case ECPGt_short:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -567,7 +586,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_int:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -587,7 +606,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_unsigned_short:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -607,7 +626,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_unsigned_int:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -627,7 +646,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_long:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -647,7 +666,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_unsigned_long:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 20, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 20, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -667,7 +686,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 #ifdef HAVE_LONG_LONG_INT_64
 			case ECPGt_long_long:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 30, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 30, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -687,7 +706,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_unsigned_long_long:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 30, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 30, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -707,7 +726,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 #endif   /* HAVE_LONG_LONG_INT_64 */
 			case ECPGt_float:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 25, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 25, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -727,7 +746,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_double:
-				if (!(mallocedval = ECPGalloc(var->arrsize * 25, stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize * 25, lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -747,7 +766,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				break;
 
 			case ECPGt_bool:
-				if (!(mallocedval = ECPGalloc(var->arrsize + sizeof("array []"), stmt->lineno)))
+				if (!(mallocedval = ECPGalloc(var->arrsize + sizeof("array []"), lineno)))
 					return false;
 
 				if (var->arrsize > 1)
@@ -766,7 +785,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 						for (element = 0; element < var->arrsize; element++)
 							sprintf(mallocedval + strlen(mallocedval), "%c,", (((int *) var->value)[element]) ? 't' : 'f');
 					else
-						ECPGraise(stmt->lineno, ECPG_CONVERT_BOOL, ECPG_SQLSTATE_DATATYPE_MISMATCH, "different size");
+						ECPGraise(lineno, ECPG_CONVERT_BOOL, ECPG_SQLSTATE_DATATYPE_MISMATCH, "different size");
 
 					strcpy(mallocedval + strlen(mallocedval) - 1, "]");
 				}
@@ -777,7 +796,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					else if (var->offset == sizeof(int))
 						sprintf(mallocedval, "'%c'", (*((int *) var->value)) ? 't' : 'f');
 					else
-						ECPGraise(stmt->lineno, ECPG_CONVERT_BOOL, ECPG_SQLSTATE_DATATYPE_MISMATCH, "different size");
+						ECPGraise(lineno, ECPG_CONVERT_BOOL, ECPG_SQLSTATE_DATATYPE_MISMATCH, "different size");
 				}
 
 				*tobeinserted_p = mallocedval;
@@ -790,16 +809,18 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					/* set slen to string length if type is char * */
 					int			slen = (var->varcharsize == 0) ? strlen((char *) var->value) : var->varcharsize;
 
-					if (!(newcopy = ECPGalloc(slen + 1, stmt->lineno)))
+					if (!(newcopy = ECPGalloc(slen + 1, lineno)))
 						return false;
 
 					strncpy(newcopy, (char *) var->value, slen);
 					newcopy[slen] = '\0';
 
-					mallocedval = quote_postgres(newcopy, stmt->lineno);
+					mallocedval = quote_postgres(newcopy, lineno);
 					if (!mallocedval)
 						return false;
 
+					ECPGfree(newcopy);
+
 					*tobeinserted_p = mallocedval;
 					*malloced_p = true;
 				}
@@ -809,7 +830,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				{
 					int			slen = strlen((char *) var->value);
 
-					if (!(mallocedval = ECPGalloc(slen + 1, stmt->lineno)))
+					if (!(mallocedval = ECPGalloc(slen + 1, lineno)))
 						return false;
 
 					strncpy(mallocedval, (char *) var->value, slen);
@@ -824,16 +845,18 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					struct ECPGgeneric_varchar *variable =
 					(struct ECPGgeneric_varchar *) (var->value);
 
-					if (!(newcopy = (char *) ECPGalloc(variable->len + 1, stmt->lineno)))
+					if (!(newcopy = (char *) ECPGalloc(variable->len + 1, lineno)))
 						return false;
 
 					strncpy(newcopy, variable->arr, variable->len);
 					newcopy[variable->len] = '\0';
 
-					mallocedval = quote_postgres(newcopy, stmt->lineno);
+					mallocedval = quote_postgres(newcopy, lineno);
 					if (!mallocedval)
 						return false;
 
+					ECPGfree(newcopy);
+
 					*tobeinserted_p = mallocedval;
 					*malloced_p = true;
 				}
@@ -848,7 +871,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 
 					if (var->arrsize > 1)
 					{
-						for (element = 0; element < var->arrsize; element++, nval = PGTYPESnumeric_new())
+						for (element = 0; element < var->arrsize; element++)
 						{
 							if (var->type == ECPGt_numeric)
 								PGTYPESnumeric_copy((numeric *) ((var + var->offset * element)->value), nval);
@@ -859,7 +882,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 							PGTYPESnumeric_free(nval);
 							slen = strlen(str);
 
-							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [] "), stmt->lineno)))
+							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [] "), lineno)))
 								return false;
 
 							if (!element)
@@ -882,7 +905,7 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 						PGTYPESnumeric_free(nval);
 						slen = strlen(str);
 
-						if (!(mallocedval = ECPGalloc(slen + 1, stmt->lineno)))
+						if (!(mallocedval = ECPGalloc(slen + 1, lineno)))
 							return false;
 
 						strncpy(mallocedval, str, slen);
@@ -904,10 +927,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					{
 						for (element = 0; element < var->arrsize; element++)
 						{
-							str = quote_postgres(PGTYPESinterval_to_asc((interval *) ((var + var->offset * element)->value)), stmt->lineno);
+							str = quote_postgres(PGTYPESinterval_to_asc((interval *) ((var + var->offset * element)->value)), lineno);
 							slen = strlen(str);
 
-							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [],interval "), stmt->lineno)))
+							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [],interval "), lineno)))
 								return false;
 
 							if (!element)
@@ -921,10 +944,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					}
 					else
 					{
-						str = quote_postgres(PGTYPESinterval_to_asc((interval *) (var->value)), stmt->lineno);
+						str = quote_postgres(PGTYPESinterval_to_asc((interval *) (var->value)), lineno);
 						slen = strlen(str);
 
-						if (!(mallocedval = ECPGalloc(slen + sizeof("interval ") + 1, stmt->lineno)))
+						if (!(mallocedval = ECPGalloc(slen + sizeof("interval ") + 1, lineno)))
 							return false;
 
 						strcpy(mallocedval, "interval ");
@@ -947,10 +970,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					{
 						for (element = 0; element < var->arrsize; element++)
 						{
-							str = quote_postgres(PGTYPESdate_to_asc(*(date *) ((var + var->offset * element)->value)), stmt->lineno);
+							str = quote_postgres(PGTYPESdate_to_asc(*(date *) ((var + var->offset * element)->value)), lineno);
 							slen = strlen(str);
 
-							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [],date "), stmt->lineno)))
+							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [],date "), lineno)))
 								return false;
 
 							if (!element)
@@ -964,10 +987,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					}
 					else
 					{
-						str = quote_postgres(PGTYPESdate_to_asc(*(date *) (var->value)), stmt->lineno);
+						str = quote_postgres(PGTYPESdate_to_asc(*(date *) (var->value)), lineno);
 						slen = strlen(str);
 
-						if (!(mallocedval = ECPGalloc(slen + sizeof("date ") + 1, stmt->lineno)))
+						if (!(mallocedval = ECPGalloc(slen + sizeof("date ") + 1, lineno)))
 							return false;
 
 						strcpy(mallocedval, "date ");
@@ -990,10 +1013,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					{
 						for (element = 0; element < var->arrsize; element++)
 						{
-							str = quote_postgres(PGTYPEStimestamp_to_asc(*(timestamp *) ((var + var->offset * element)->value)), stmt->lineno);
+							str = quote_postgres(PGTYPEStimestamp_to_asc(*(timestamp *) ((var + var->offset * element)->value)), lineno);
 							slen = strlen(str);
 
-							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [], timestamp "), stmt->lineno)))
+							if (!(mallocedval = ECPGrealloc(mallocedval, strlen(mallocedval) + slen + sizeof("array [], timestamp "), lineno)))
 								return false;
 
 							if (!element)
@@ -1007,10 +1030,10 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 					}
 					else
 					{
-						str = quote_postgres(PGTYPEStimestamp_to_asc(*(timestamp *) (var->value)), stmt->lineno);
+						str = quote_postgres(PGTYPEStimestamp_to_asc(*(timestamp *) (var->value)), lineno);
 						slen = strlen(str);
 
-						if (!(mallocedval = ECPGalloc(slen + sizeof("timestamp") + 1, stmt->lineno)))
+						if (!(mallocedval = ECPGalloc(slen + sizeof("timestamp") + 1, lineno)))
 							return false;
 
 						strcpy(mallocedval, "timestamp ");
@@ -1024,9 +1047,12 @@ ECPGstore_input(const struct statement * stmt, const struct variable * var,
 				}
 				break;
 
+			case ECPGt_descriptor:
+				break;
+
 			default:
 				/* Not implemented yet */
-				ECPGraise(stmt->lineno, ECPG_UNSUPPORTED, ECPG_SQLSTATE_ECPG_INTERNAL_ERROR, (char *) ECPGtype_name(var->type));
+				ECPGraise(lineno, ECPG_UNSUPPORTED, ECPG_SQLSTATE_ECPG_INTERNAL_ERROR, (char *) ECPGtype_name(var->type));
 				return false;
 				break;
 		}
@@ -1043,6 +1069,7 @@ ECPGexecute(struct statement * stmt)
 	PGresult   *results;
 	PGnotify   *notify;
 	struct variable *var;
+	int			desc_counter = 0;
 
 	copiedquery = ECPGstrdup(stmt->command, stmt->lineno);
 
@@ -1053,64 +1080,140 @@ ECPGexecute(struct statement * stmt)
 	 * so on.
 	 */
 	var = stmt->inlist;
+
 	while (var)
 	{
 		char	   *newcopy = NULL;
-		const char *tobeinserted = NULL;
+		const char *tobeinserted;
 		char	   *p;
 		bool		malloced = FALSE;
 		int			hostvarl = 0;
 
-		if (!ECPGstore_input(stmt, var, &tobeinserted, &malloced))
-			return false;
+		tobeinserted = NULL;
 
 		/*
-		 * Now tobeinserted points to an area that is to be inserted at
-		 * the first %s
+		 * A descriptor is a special case since it contains many variables
+		 * but is listed only once.
 		 */
-		if (!(newcopy = (char *) ECPGalloc(strlen(copiedquery) + strlen(tobeinserted) + 1, stmt->lineno)))
-			return false;
-
-		strcpy(newcopy, copiedquery);
-		if ((p = next_insert(newcopy + hostvarl)) == NULL)
+		if (var->type == ECPGt_descriptor)
 		{
 			/*
-			 * We have an argument but we dont have the matched up string
-			 * in the string
+			 * We create an additional variable list here, so the same
+			 * logic applies.
 			 */
-			ECPGraise(stmt->lineno, ECPG_TOO_MANY_ARGUMENTS, ECPG_SQLSTATE_USING_CLAUSE_DOES_NOT_MATCH_PARAMETERS, NULL);
-			return false;
+			struct variable desc_inlist;
+			struct descriptor *desc;
+			struct descriptor_item *desc_item;
+
+			for (desc = all_descriptors; desc; desc = desc->next)
+			{
+				if (strcmp(var->pointer, desc->name) == 0)
+					break;
+			}
+
+			if (desc == NULL)
+			{
+				ECPGraise(stmt->lineno, ECPG_UNKNOWN_DESCRIPTOR, ECPG_SQLSTATE_INVALID_SQL_DESCRIPTOR_NAME, var->pointer);
+				return false;
+			}
+
+			desc_counter++;
+			if (desc->count < 0 || desc->count >= desc_counter)
+			{
+				for (desc_item = desc->items; desc_item; desc_item = desc_item->next)
+				{
+					if (desc_item->num == desc_counter)
+					{
+						desc_inlist.type = ECPGt_char;
+						desc_inlist.value = desc_item->data;
+						desc_inlist.pointer = &(desc_item->data);
+						desc_inlist.varcharsize = strlen(desc_item->data);
+						desc_inlist.arrsize = 1;
+						desc_inlist.offset = 0;
+						if (!desc_item->indicator)
+						{
+							desc_inlist.ind_type = ECPGt_NO_INDICATOR;
+							desc_inlist.ind_value = desc_inlist.ind_pointer = NULL;
+							desc_inlist.ind_varcharsize = desc_inlist.ind_arrsize = desc_inlist.ind_offset = 0;
+						}
+						else
+						{
+							desc_inlist.ind_type = ECPGt_int;
+							desc_inlist.ind_value = &(desc_item->indicator);
+							desc_inlist.ind_pointer = &(desc_inlist.ind_value);
+							desc_inlist.ind_varcharsize = desc_inlist.ind_arrsize = 1;
+							desc_inlist.ind_offset = 0;
+						}
+						if (!ECPGstore_input(stmt->lineno, stmt->force_indicator, &desc_inlist, &tobeinserted, &malloced))
+							return false;
+
+						break;
+					}
+				}
+
+				if (!desc_item) /* no more entries found in descriptor */
+					desc_counter = 0;
+			}
+			else
+				desc_counter = 0;
 		}
 		else
 		{
-			strcpy(p, tobeinserted);
-			hostvarl = strlen(newcopy);
+			if (!ECPGstore_input(stmt->lineno, stmt->force_indicator, var, &tobeinserted, &malloced))
+				return false;
+		}
+
+		if (tobeinserted)
+		{
+			/*
+			 * Now tobeinserted points to an area that is to be inserted
+			 * at the first %s
+			 */
+			if (!(newcopy = (char *) ECPGalloc(strlen(copiedquery) + strlen(tobeinserted) + 1, stmt->lineno)))
+				return false;
+
+			strcpy(newcopy, copiedquery);
+			if ((p = next_insert(newcopy + hostvarl)) == NULL)
+			{
+				/*
+				 * We have an argument but we dont have the matched up
+				 * string in the string
+				 */
+				ECPGraise(stmt->lineno, ECPG_TOO_MANY_ARGUMENTS, ECPG_SQLSTATE_USING_CLAUSE_DOES_NOT_MATCH_PARAMETERS, NULL);
+				return false;
+			}
+			else
+			{
+				strcpy(p, tobeinserted);
+				hostvarl = strlen(newcopy);
+
+				/*
+				 * The strange thing in the second argument is the rest of
+				 * the string from the old string
+				 */
+				strcat(newcopy,
+					   copiedquery
+					   + (p - newcopy)
+					   + sizeof("?") - 1 /* don't count the '\0' */ );
+			}
 
 			/*
-			 * The strange thing in the second argument is the rest of the
-			 * string from the old string
+			 * Now everything is safely copied to the newcopy. Lets free
+			 * the oldcopy and let the copiedquery get the var->value from
+			 * the newcopy.
 			 */
-			strcat(newcopy,
-				   copiedquery
-				   + (p - newcopy)
-				   + sizeof("?") - 1 /* don't count the '\0' */ );
+			if (malloced)
+			{
+				ECPGfree((char *) tobeinserted);
+				tobeinserted = NULL;
+			}
+
+			ECPGfree(copiedquery);
+			copiedquery = newcopy;
 		}
 
-		/*
-		 * Now everything is safely copied to the newcopy. Lets free the
-		 * oldcopy and let the copiedquery get the var->value from the
-		 * newcopy.
-		 */
-		if (malloced)
-		{
-			ECPGfree((char *) tobeinserted);
-			tobeinserted = NULL;
-		}
-
-		ECPGfree(copiedquery);
-		copiedquery = newcopy;
-
-		var = var->next;
+		if (desc_counter == 0)
+			var = var->next;
 	}
 
 	/* Check if there are unmatched things left. */
