@@ -16,16 +16,14 @@
  * will cause repeat printouts.
  *
  * We must speak the same transmitted data representations as the backend
- * routines.  Note that this module supports *only* network byte order
- * for transmitted ints, whereas the backend modules (as of this writing)
- * still handle either network or little-endian byte order.
+ * routines.
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ *
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-misc.c,v 1.65 2001/12/03 00:28:24 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/interfaces/libpq/fe-misc.c,v 1.112 2004/12/31 22:03:50 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -36,15 +34,24 @@
 #include <signal.h>
 #include <time.h>
 
+#ifndef WIN32_CLIENT_ONLY
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#endif
+
 #ifdef WIN32
-#define WIN32_LEAN_AND_MEAN
-#include <windows.h>
 #include "win32.h"
 #else
 #include <unistd.h>
 #include <sys/time.h>
 #endif
 
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#endif
+#ifdef HAVE_SYS_POLL_H
+#include <sys/poll.h>
+#endif
 #ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
@@ -52,21 +59,18 @@
 #include "libpq-fe.h"
 #include "libpq-int.h"
 #include "pqsignal.h"
-
-#ifdef MULTIBYTE
 #include "mb/pg_wchar.h"
-#endif
 
 
-#define DONOTICE(conn,message) \
-	((*(conn)->noticeHook) ((conn)->noticeArg, (message)))
-
-static int	pqPutBytes(const char *s, size_t nbytes, PGconn *conn);
+static int	pqPutMsgBytes(const void *buf, size_t len, PGconn *conn);
+static int	pqSendSome(PGconn *conn, int len);
+static int pqSocketCheck(PGconn *conn, int forRead, int forWrite,
+			  time_t end_time);
+static int	pqSocketPoll(int sock, int forRead, int forWrite, time_t end_time);
 
 
 /*
- * pqGetc:
- *	get a character from the connection
+ * pqGetc: get 1 character from the connection
  *
  *	All these routines return 0 on success, EOF on error.
  *	Note that for the Get routines, EOF only means there is not enough
@@ -88,12 +92,12 @@ pqGetc(char *result, PGconn *conn)
 
 
 /*
- * write 1 char to the connection
+ * pqPutc: write 1 char to the current message
  */
 int
 pqPutc(char c, PGconn *conn)
 {
-	if (pqPutBytes(&c, 1, conn) == EOF)
+	if (pqPutMsgBytes(&c, 1, conn))
 		return EOF;
 
 	if (conn->Pfdebug)
@@ -103,65 +107,6 @@ pqPutc(char c, PGconn *conn)
 }
 
 
-/*
- * pqPutBytes: local routine to write N bytes to the connection,
- * with buffering
- */
-static int
-pqPutBytes(const char *s, size_t nbytes, PGconn *conn)
-{
-	size_t		avail = Max(conn->outBufSize - conn->outCount, 0);
-
-	/*
-	 * if we are non-blocking and the send queue is too full to buffer
-	 * this request then try to flush some and return an error
-	 */
-	if (pqIsnonblocking(conn) && nbytes > avail && pqFlush(conn))
-	{
-		/*
-		 * even if the flush failed we may still have written some data,
-		 * recalculate the size of the send-queue relative to the amount
-		 * we have to send, we may be able to queue it afterall even
-		 * though it's not sent to the database it's ok, any routines that
-		 * check the data coming from the database better call pqFlush()
-		 * anyway.
-		 */
-		if (nbytes > Max(conn->outBufSize - conn->outCount, 0))
-		{
-			printfPQExpBuffer(&conn->errorMessage,
-							  libpq_gettext("could not flush enough data (space available: %d, space needed %d)\n"),
-						 (int) Max(conn->outBufSize - conn->outCount, 0),
-							  (int) nbytes);
-			return EOF;
-		}
-		/* fixup avail for while loop */
-		avail = Max(conn->outBufSize - conn->outCount, 0);
-	}
-
-	/*
-	 * is the amount of data to be sent is larger than the size of the
-	 * output buffer then we must flush it to make more room.
-	 *
-	 * the code above will make sure the loop conditional is never true for
-	 * non-blocking connections
-	 */
-	while (nbytes > avail)
-	{
-		memcpy(conn->outBuffer + conn->outCount, s, avail);
-		conn->outCount += avail;
-		s += avail;
-		nbytes -= avail;
-		if (pqFlush(conn))
-			return EOF;
-		avail = conn->outBufSize;
-	}
-
-	memcpy(conn->outBuffer + conn->outCount, s, nbytes);
-	conn->outCount += nbytes;
-
-	return 0;
-}
-
 /*
  * pqGets:
  * get a null-terminated string from the connection,
@@ -199,14 +144,17 @@ pqGets(PQExpBuffer buf, PGconn *conn)
 }
 
 
+/*
+ * pqPuts: write a null-terminated string to the current message
+ */
 int
 pqPuts(const char *s, PGconn *conn)
 {
-	if (pqPutBytes(s, strlen(s) + 1, conn))
+	if (pqPutMsgBytes(s, strlen(s) + 1, conn))
 		return EOF;
 
 	if (conn->Pfdebug)
-		fprintf(conn->Pfdebug, "To backend> %s\n", s);
+		fprintf(conn->Pfdebug, "To backend> '%s'\n", s);
 
 	return 0;
 }
@@ -218,7 +166,7 @@ pqPuts(const char *s, PGconn *conn)
 int
 pqGetnchar(char *s, size_t len, PGconn *conn)
 {
-	if (len < 0 || len > conn->inEnd - conn->inCursor)
+	if (len < 0 || len > (size_t) (conn->inEnd - conn->inCursor))
 		return EOF;
 
 	memcpy(s, conn->inBuffer + conn->inCursor, len);
@@ -234,12 +182,12 @@ pqGetnchar(char *s, size_t len, PGconn *conn)
 
 /*
  * pqPutnchar:
- *	send a string of exactly len bytes, no null termination needed
+ *	write exactly len bytes to the current message
  */
 int
 pqPutnchar(const char *s, size_t len, PGconn *conn)
 {
-	if (pqPutBytes(s, len, conn))
+	if (pqPutMsgBytes(s, len, conn))
 		return EOF;
 
 	if (conn->Pfdebug)
@@ -249,7 +197,7 @@ pqPutnchar(const char *s, size_t len, PGconn *conn)
 }
 
 /*
- * pgGetInt
+ * pqGetInt
  *	read a 2 or 4 byte integer and convert from network byte order
  *	to local byte order
  */
@@ -258,7 +206,6 @@ pqGetInt(int *result, size_t bytes, PGconn *conn)
 {
 	uint16		tmp2;
 	uint32		tmp4;
-	char		noticeBuf[64];
 
 	switch (bytes)
 	{
@@ -277,10 +224,9 @@ pqGetInt(int *result, size_t bytes, PGconn *conn)
 			*result = (int) ntohl(tmp4);
 			break;
 		default:
-			snprintf(noticeBuf, sizeof(noticeBuf),
-					 libpq_gettext("integer of size %lu not supported by pqGetInt\n"),
-					 (unsigned long) bytes);
-			DONOTICE(conn, noticeBuf);
+			pqInternalNotice(&conn->noticeHooks,
+						 "integer of size %lu not supported by pqGetInt",
+							 (unsigned long) bytes);
 			return EOF;
 	}
 
@@ -291,8 +237,8 @@ pqGetInt(int *result, size_t bytes, PGconn *conn)
 }
 
 /*
- * pgPutInt
- * send an integer of 2 or 4 bytes, converting from host byte order
+ * pqPutInt
+ * write an integer of 2 or 4 bytes, converting from host byte order
  * to network byte order.
  */
 int
@@ -300,25 +246,23 @@ pqPutInt(int value, size_t bytes, PGconn *conn)
 {
 	uint16		tmp2;
 	uint32		tmp4;
-	char		noticeBuf[64];
 
 	switch (bytes)
 	{
 		case 2:
 			tmp2 = htons((uint16) value);
-			if (pqPutBytes((const char *) &tmp2, 2, conn))
+			if (pqPutMsgBytes((const char *) &tmp2, 2, conn))
 				return EOF;
 			break;
 		case 4:
 			tmp4 = htonl((uint32) value);
-			if (pqPutBytes((const char *) &tmp4, 4, conn))
+			if (pqPutMsgBytes((const char *) &tmp4, 4, conn))
 				return EOF;
 			break;
 		default:
-			snprintf(noticeBuf, sizeof(noticeBuf),
-					 libpq_gettext("integer of size %lu not supported by pqPutInt\n"),
-					 (unsigned long) bytes);
-			DONOTICE(conn, noticeBuf);
+			pqInternalNotice(&conn->noticeHooks,
+						 "integer of size %lu not supported by pqPutInt",
+							 (unsigned long) bytes);
 			return EOF;
 	}
 
@@ -329,70 +273,251 @@ pqPutInt(int value, size_t bytes, PGconn *conn)
 }
 
 /*
- * pqReadReady: is select() saying the file is ready to read?
- * Returns -1 on failure, 0 if not ready, 1 if ready.
+ * Make sure conn's output buffer can hold bytes_needed bytes (caller must
+ * include already-stored data into the value!)
+ *
+ * Returns 0 on success, EOF if failed to enlarge buffer
  */
 int
-pqReadReady(PGconn *conn)
+pqCheckOutBufferSpace(int bytes_needed, PGconn *conn)
 {
-	fd_set		input_mask;
-	struct timeval timeout;
+	int			newsize = conn->outBufSize;
+	char	   *newbuf;
 
-	if (!conn || conn->sock < 0)
-		return -1;
+	if (bytes_needed <= newsize)
+		return 0;
 
-retry:
-	FD_ZERO(&input_mask);
-	FD_SET(conn->sock, &input_mask);
-	timeout.tv_sec = 0;
-	timeout.tv_usec = 0;
-	if (select(conn->sock + 1, &input_mask, (fd_set *) NULL, (fd_set *) NULL,
-			   &timeout) < 0)
+	/*
+	 * If we need to enlarge the buffer, we first try to double it in
+	 * size; if that doesn't work, enlarge in multiples of 8K.  This
+	 * avoids thrashing the malloc pool by repeated small enlargements.
+	 *
+	 * Note: tests for newsize > 0 are to catch integer overflow.
+	 */
+	do
 	{
-		if (SOCK_ERRNO == EINTR)
-			/* Interrupted system call - we'll just try again */
-			goto retry;
+		newsize *= 2;
+	} while (bytes_needed > newsize && newsize > 0);
 
-		printfPQExpBuffer(&conn->errorMessage,
-						  libpq_gettext("select() failed: %s\n"),
-						  SOCK_STRERROR(SOCK_ERRNO));
-		return -1;
+	if (bytes_needed <= newsize)
+	{
+		newbuf = realloc(conn->outBuffer, newsize);
+		if (newbuf)
+		{
+			/* realloc succeeded */
+			conn->outBuffer = newbuf;
+			conn->outBufSize = newsize;
+			return 0;
+		}
 	}
 
-	return FD_ISSET(conn->sock, &input_mask) ? 1 : 0;
+	newsize = conn->outBufSize;
+	do
+	{
+		newsize += 8192;
+	} while (bytes_needed > newsize && newsize > 0);
+
+	if (bytes_needed <= newsize)
+	{
+		newbuf = realloc(conn->outBuffer, newsize);
+		if (newbuf)
+		{
+			/* realloc succeeded */
+			conn->outBuffer = newbuf;
+			conn->outBufSize = newsize;
+			return 0;
+		}
+	}
+
+	/* realloc failed. Probably out of memory */
+	printfPQExpBuffer(&conn->errorMessage,
+					  "cannot allocate memory for output buffer\n");
+	return EOF;
 }
 
 /*
- * pqWriteReady: is select() saying the file is ready to write?
- * Returns -1 on failure, 0 if not ready, 1 if ready.
+ * Make sure conn's input buffer can hold bytes_needed bytes (caller must
+ * include already-stored data into the value!)
+ *
+ * Returns 0 on success, EOF if failed to enlarge buffer
  */
 int
-pqWriteReady(PGconn *conn)
+pqCheckInBufferSpace(int bytes_needed, PGconn *conn)
 {
-	fd_set		input_mask;
-	struct timeval timeout;
+	int			newsize = conn->inBufSize;
+	char	   *newbuf;
 
-	if (!conn || conn->sock < 0)
-		return -1;
+	if (bytes_needed <= newsize)
+		return 0;
 
-retry:
-	FD_ZERO(&input_mask);
-	FD_SET(conn->sock, &input_mask);
-	timeout.tv_sec = 0;
-	timeout.tv_usec = 0;
-	if (select(conn->sock + 1, (fd_set *) NULL, &input_mask, (fd_set *) NULL,
-			   &timeout) < 0)
+	/*
+	 * If we need to enlarge the buffer, we first try to double it in
+	 * size; if that doesn't work, enlarge in multiples of 8K.  This
+	 * avoids thrashing the malloc pool by repeated small enlargements.
+	 *
+	 * Note: tests for newsize > 0 are to catch integer overflow.
+	 */
+	do
 	{
-		if (SOCK_ERRNO == EINTR)
-			/* Interrupted system call - we'll just try again */
-			goto retry;
+		newsize *= 2;
+	} while (bytes_needed > newsize && newsize > 0);
 
-		printfPQExpBuffer(&conn->errorMessage,
-						  libpq_gettext("select() failed: %s\n"),
-						  SOCK_STRERROR(SOCK_ERRNO));
-		return -1;
+	if (bytes_needed <= newsize)
+	{
+		newbuf = realloc(conn->inBuffer, newsize);
+		if (newbuf)
+		{
+			/* realloc succeeded */
+			conn->inBuffer = newbuf;
+			conn->inBufSize = newsize;
+			return 0;
+		}
 	}
-	return FD_ISSET(conn->sock, &input_mask) ? 1 : 0;
+
+	newsize = conn->inBufSize;
+	do
+	{
+		newsize += 8192;
+	} while (bytes_needed > newsize && newsize > 0);
+
+	if (bytes_needed <= newsize)
+	{
+		newbuf = realloc(conn->inBuffer, newsize);
+		if (newbuf)
+		{
+			/* realloc succeeded */
+			conn->inBuffer = newbuf;
+			conn->inBufSize = newsize;
+			return 0;
+		}
+	}
+
+	/* realloc failed. Probably out of memory */
+	printfPQExpBuffer(&conn->errorMessage,
+					  "cannot allocate memory for input buffer\n");
+	return EOF;
+}
+
+/*
+ * pqPutMsgStart: begin construction of a message to the server
+ *
+ * msg_type is the message type byte, or 0 for a message without type byte
+ * (only startup messages have no type byte)
+ *
+ * force_len forces the message to have a length word; otherwise, we add
+ * a length word if protocol 3.
+ *
+ * Returns 0 on success, EOF on error
+ *
+ * The idea here is that we construct the message in conn->outBuffer,
+ * beginning just past any data already in outBuffer (ie, at
+ * outBuffer+outCount).  We enlarge the buffer as needed to hold the message.
+ * When the message is complete, we fill in the length word (if needed) and
+ * then advance outCount past the message, making it eligible to send.
+ *
+ * The state variable conn->outMsgStart points to the incomplete message's
+ * length word: it is either outCount or outCount+1 depending on whether
+ * there is a type byte.  If we are sending a message without length word
+ * (pre protocol 3.0 only), then outMsgStart is -1.  The state variable
+ * conn->outMsgEnd is the end of the data collected so far.
+ */
+int
+pqPutMsgStart(char msg_type, bool force_len, PGconn *conn)
+{
+	int			lenPos;
+	int			endPos;
+
+	/* allow room for message type byte */
+	if (msg_type)
+		endPos = conn->outCount + 1;
+	else
+		endPos = conn->outCount;
+
+	/* do we want a length word? */
+	if (force_len || PG_PROTOCOL_MAJOR(conn->pversion) >= 3)
+	{
+		lenPos = endPos;
+		/* allow room for message length */
+		endPos += 4;
+	}
+	else
+		lenPos = -1;
+
+	/* make sure there is room for message header */
+	if (pqCheckOutBufferSpace(endPos, conn))
+		return EOF;
+	/* okay, save the message type byte if any */
+	if (msg_type)
+		conn->outBuffer[conn->outCount] = msg_type;
+	/* set up the message pointers */
+	conn->outMsgStart = lenPos;
+	conn->outMsgEnd = endPos;
+	/* length word, if needed, will be filled in by pqPutMsgEnd */
+
+	if (conn->Pfdebug)
+		fprintf(conn->Pfdebug, "To backend> Msg %c\n",
+				msg_type ? msg_type : ' ');
+
+	return 0;
+}
+
+/*
+ * pqPutMsgBytes: add bytes to a partially-constructed message
+ *
+ * Returns 0 on success, EOF on error
+ */
+static int
+pqPutMsgBytes(const void *buf, size_t len, PGconn *conn)
+{
+	/* make sure there is room for it */
+	if (pqCheckOutBufferSpace(conn->outMsgEnd + len, conn))
+		return EOF;
+	/* okay, save the data */
+	memcpy(conn->outBuffer + conn->outMsgEnd, buf, len);
+	conn->outMsgEnd += len;
+	/* no Pfdebug call here, caller should do it */
+	return 0;
+}
+
+/*
+ * pqPutMsgEnd: finish constructing a message and possibly send it
+ *
+ * Returns 0 on success, EOF on error
+ *
+ * We don't actually send anything here unless we've accumulated at least
+ * 8K worth of data (the typical size of a pipe buffer on Unix systems).
+ * This avoids sending small partial packets.  The caller must use pqFlush
+ * when it's important to flush all the data out to the server.
+ */
+int
+pqPutMsgEnd(PGconn *conn)
+{
+	if (conn->Pfdebug)
+		fprintf(conn->Pfdebug, "To backend> Msg complete, length %u\n",
+				conn->outMsgEnd - conn->outCount);
+
+	/* Fill in length word if needed */
+	if (conn->outMsgStart >= 0)
+	{
+		uint32		msgLen = conn->outMsgEnd - conn->outMsgStart;
+
+		msgLen = htonl(msgLen);
+		memcpy(conn->outBuffer + conn->outMsgStart, &msgLen, 4);
+	}
+
+	/* Make message eligible to send */
+	conn->outCount = conn->outMsgEnd;
+
+	if (conn->outCount >= 8192)
+	{
+		int			toSend = conn->outCount - (conn->outCount % 8192);
+
+		if (pqSendSome(conn, toSend) < 0)
+			return EOF;
+		/* in nonblock mode, don't complain if unable to send it all */
+	}
+
+	return 0;
 }
 
 /* ----------
@@ -411,6 +536,7 @@ pqReadData(PGconn *conn)
 {
 	int			someread = 0;
 	int			nread;
+	char		sebuf[256];
 
 	if (conn->sock < 0)
 	{
@@ -447,30 +573,25 @@ pqReadData(PGconn *conn)
 	 */
 	if (conn->inBufSize - conn->inEnd < 8192)
 	{
-		int			newSize = conn->inBufSize * 2;
-		char	   *newBuf = (char *) realloc(conn->inBuffer, newSize);
-
-		if (newBuf)
+		if (pqCheckInBufferSpace(conn->inEnd + 8192, conn))
 		{
-			conn->inBuffer = newBuf;
-			conn->inBufSize = newSize;
+			/*
+			 * We don't insist that the enlarge worked, but we need some
+			 * room
+			 */
+			if (conn->inBufSize - conn->inEnd < 100)
+				return -1;		/* errorMessage already set */
 		}
 	}
 
 	/* OK, try to read some data */
-tryAgain:
-#ifdef USE_SSL
-	if (conn->ssl)
-		nread = SSL_read(conn->ssl, conn->inBuffer + conn->inEnd,
-						 conn->inBufSize - conn->inEnd);
-	else
-#endif
-		nread = recv(conn->sock, conn->inBuffer + conn->inEnd,
-					 conn->inBufSize - conn->inEnd, 0);
+retry3:
+	nread = pqsecure_read(conn, conn->inBuffer + conn->inEnd,
+						  conn->inBufSize - conn->inEnd);
 	if (nread < 0)
 	{
 		if (SOCK_ERRNO == EINTR)
-			goto tryAgain;
+			goto retry3;
 		/* Some systems return EAGAIN/EWOULDBLOCK for no data */
 #ifdef EAGAIN
 		if (SOCK_ERRNO == EAGAIN)
@@ -487,7 +608,7 @@ tryAgain:
 #endif
 		printfPQExpBuffer(&conn->errorMessage,
 			   libpq_gettext("could not receive data from server: %s\n"),
-						  SOCK_STRERROR(SOCK_ERRNO));
+						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
 		return -1;
 	}
 	if (nread > 0)
@@ -511,7 +632,7 @@ tryAgain:
 			(conn->inBufSize - conn->inEnd) >= 8192)
 		{
 			someread = 1;
-			goto tryAgain;
+			goto retry3;
 		}
 		return 1;
 	}
@@ -527,7 +648,18 @@ tryAgain:
 	 * file is ready. Grumble.	Fortunately, we don't expect this path to
 	 * be taken much, since in normal practice we should not be trying to
 	 * read data unless the file selected for reading already.
+	 *
+	 * In SSL mode it's even worse: SSL_read() could say WANT_READ and then
+	 * data could arrive before we make the pqReadReady() test.  So we
+	 * must play dumb and assume there is more data, relying on the SSL
+	 * layer to detect true EOF.
 	 */
+
+#ifdef USE_SSL
+	if (conn->ssl)
+		return 0;
+#endif
+
 	switch (pqReadReady(conn))
 	{
 		case 0:
@@ -544,19 +676,13 @@ tryAgain:
 	 * Still not sure that it's EOF, because some data could have just
 	 * arrived.
 	 */
-tryAgain2:
-#ifdef USE_SSL
-	if (conn->ssl)
-		nread = SSL_read(conn->ssl, conn->inBuffer + conn->inEnd,
-						 conn->inBufSize - conn->inEnd);
-	else
-#endif
-		nread = recv(conn->sock, conn->inBuffer + conn->inEnd,
-					 conn->inBufSize - conn->inEnd, 0);
+retry4:
+	nread = pqsecure_read(conn, conn->inBuffer + conn->inEnd,
+						  conn->inBufSize - conn->inEnd);
 	if (nread < 0)
 	{
 		if (SOCK_ERRNO == EINTR)
-			goto tryAgain2;
+			goto retry4;
 		/* Some systems return EAGAIN/EWOULDBLOCK for no data */
 #ifdef EAGAIN
 		if (SOCK_ERRNO == EAGAIN)
@@ -573,7 +699,7 @@ tryAgain2:
 #endif
 		printfPQExpBuffer(&conn->errorMessage,
 			   libpq_gettext("could not receive data from server: %s\n"),
-						  SOCK_STRERROR(SOCK_ERRNO));
+						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
 		return -1;
 	}
 	if (nread > 0)
@@ -593,65 +719,49 @@ definitelyFailed:
 			   "\tThis probably means the server terminated abnormally\n"
 						 "\tbefore or while processing the request.\n"));
 	conn->status = CONNECTION_BAD;		/* No more connection to backend */
-#ifdef WIN32
+	pqsecure_close(conn);
 	closesocket(conn->sock);
-#else
-	close(conn->sock);
-#endif
 	conn->sock = -1;
 
 	return -1;
 }
 
 /*
- * pqFlush: send any data waiting in the output buffer
+ * pqSendSome: send data waiting in the output buffer.
+ *
+ * len is how much to try to send (typically equal to outCount, but may
+ * be less).
+ *
+ * Return 0 on success, -1 on failure and 1 when not all data could be sent
+ * because the socket would block and the connection is non-blocking.
  */
-int
-pqFlush(PGconn *conn)
+static int
+pqSendSome(PGconn *conn, int len)
 {
 	char	   *ptr = conn->outBuffer;
-	int			len = conn->outCount;
+	int			remaining = conn->outCount;
+	int			result = 0;
 
 	if (conn->sock < 0)
 	{
 		printfPQExpBuffer(&conn->errorMessage,
 						  libpq_gettext("connection not open\n"));
-		return EOF;
+		return -1;
 	}
 
-	/*
-	 * don't try to send zero data, allows us to use this function without
-	 * too much worry about overhead
-	 */
-	if (len == 0)
-		return (0);
-
 	/* while there's still data to send */
 	while (len > 0)
 	{
-		/* Prevent being SIGPIPEd if backend has closed the connection. */
-#ifndef WIN32
-		pqsigfunc	oldsighandler = pqsignal(SIGPIPE, SIG_IGN);
-#endif
-
 		int			sent;
+		char		sebuf[256];
 
-#ifdef USE_SSL
-		if (conn->ssl)
-			sent = SSL_write(conn->ssl, ptr, len);
-		else
-#endif
-			sent = send(conn->sock, ptr, len, 0);
-
-#ifndef WIN32
-		pqsignal(SIGPIPE, oldsighandler);
-#endif
+		sent = pqsecure_write(conn, ptr, len);
 
 		if (sent < 0)
 		{
 			/*
-			 * Anything except EAGAIN or EWOULDBLOCK is trouble. If it's
-			 * EPIPE or ECONNRESET, assume we've lost the backend
+			 * Anything except EAGAIN/EWOULDBLOCK/EINTR is trouble. If
+			 * it's EPIPE or ECONNRESET, assume we've lost the backend
 			 * connection permanently.
 			 */
 			switch (SOCK_ERRNO)
@@ -683,64 +793,101 @@ pqFlush(PGconn *conn)
 					 * (typically, a NOTICE message from the backend
 					 * telling us it's committing hara-kiri...).  Leave
 					 * the socket open until pqReadData finds no more data
-					 * can be read.
+					 * can be read.  But abandon attempt to send data.
 					 */
-					return EOF;
+					conn->outCount = 0;
+					return -1;
 
 				default:
 					printfPQExpBuffer(&conn->errorMessage,
 					libpq_gettext("could not send data to server: %s\n"),
-									  SOCK_STRERROR(SOCK_ERRNO));
+						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
 					/* We don't assume it's a fatal error... */
-					return EOF;
+					conn->outCount = 0;
+					return -1;
 			}
 		}
 		else
 		{
 			ptr += sent;
 			len -= sent;
+			remaining -= sent;
 		}
 
 		if (len > 0)
 		{
-			/* We didn't send it all, wait till we can send more */
+			/*
+			 * We didn't send it all, wait till we can send more.
+			 *
+			 * If the connection is in non-blocking mode we don't wait, but
+			 * return 1 to indicate that data is still pending.
+			 */
+			if (pqIsnonblocking(conn))
+			{
+				result = 1;
+				break;
+			}
 
 			/*
-			 * if the socket is in non-blocking mode we may need to abort
-			 * here
+			 * There are scenarios in which we can't send data because the
+			 * communications channel is full, but we cannot expect the
+			 * server to clear the channel eventually because it's blocked
+			 * trying to send data to us.  (This can happen when we are
+			 * sending a large amount of COPY data, and the server has
+			 * generated lots of NOTICE responses.)  To avoid a deadlock
+			 * situation, we must be prepared to accept and buffer
+			 * incoming data before we try again.  Furthermore, it is
+			 * possible that such incoming data might not arrive until
+			 * after we've gone to sleep.  Therefore, we wait for either
+			 * read ready or write ready.
 			 */
-#ifdef USE_SSL
-			/* can't do anything for our SSL users yet */
-			if (conn->ssl == NULL)
+			if (pqReadData(conn) < 0)
 			{
-#endif
-				if (pqIsnonblocking(conn))
-				{
-					/* shift the contents of the buffer */
-					memmove(conn->outBuffer, ptr, len);
-					conn->outCount = len;
-					return EOF;
-				}
-#ifdef USE_SSL
+				result = -1;	/* error message already set up */
+				break;
+			}
+			if (pqWait(TRUE, TRUE, conn))
+			{
+				result = -1;
+				break;
 			}
-#endif
-
-			if (pqWait(FALSE, TRUE, conn))
-				return EOF;
 		}
 	}
 
-	conn->outCount = 0;
+	/* shift the remaining contents of the buffer */
+	if (remaining > 0)
+		memmove(conn->outBuffer, ptr, remaining);
+	conn->outCount = remaining;
 
+	return result;
+}
+
+
+/*
+ * pqFlush: send any data waiting in the output buffer
+ *
+ * Return 0 on success, -1 on failure and 1 when not all data could be sent
+ * because the socket would block and the connection is non-blocking.
+ */
+int
+pqFlush(PGconn *conn)
+{
 	if (conn->Pfdebug)
 		fflush(conn->Pfdebug);
 
+	if (conn->outCount > 0)
+		return pqSendSome(conn, conn->outCount);
+
 	return 0;
 }
 
+
 /*
  * pqWait: wait until we can read or write the connection socket
  *
+ * JAB: If SSL enabled and used and forRead, buffered bytes short-circuit the
+ * call to select().
+ *
  * We also stop waiting and return if the kernel flags an exception condition
  * on the socket.  The actual error condition will be detected and reported
  * when the caller tries to read or write the socket.
@@ -748,43 +895,190 @@ pqFlush(PGconn *conn)
 int
 pqWait(int forRead, int forWrite, PGconn *conn)
 {
-	fd_set		input_mask;
-	fd_set		output_mask;
-	fd_set		except_mask;
+	return pqWaitTimed(forRead, forWrite, conn, (time_t) -1);
+}
 
-	if (conn->sock < 0)
+/*
+ * pqWaitTimed: wait, but not past finish_time.
+ *
+ * If finish_time is exceeded then we return failure (EOF).  This is like
+ * the response for a kernel exception because we don't want the caller
+ * to try to read/write in that case.
+ *
+ * finish_time = ((time_t) -1) disables the wait limit.
+ */
+int
+pqWaitTimed(int forRead, int forWrite, PGconn *conn, time_t finish_time)
+{
+	int			result;
+
+	result = pqSocketCheck(conn, forRead, forWrite, finish_time);
+
+	if (result < 0)
+		return EOF;				/* errorMessage is already set */
+
+	if (result == 0)
 	{
 		printfPQExpBuffer(&conn->errorMessage,
-						  libpq_gettext("connection not open\n"));
+						  libpq_gettext("timeout expired\n"));
 		return EOF;
 	}
 
-	if (forRead || forWrite)
-	{
-retry:
-		FD_ZERO(&input_mask);
-		FD_ZERO(&output_mask);
-		FD_ZERO(&except_mask);
-		if (forRead)
-			FD_SET(conn->sock, &input_mask);
-		if (forWrite)
-			FD_SET(conn->sock, &output_mask);
-		FD_SET(conn->sock, &except_mask);
-		if (select(conn->sock + 1, &input_mask, &output_mask, &except_mask,
-				   (struct timeval *) NULL) < 0)
-		{
-			if (SOCK_ERRNO == EINTR)
-				goto retry;
-			printfPQExpBuffer(&conn->errorMessage,
-							  libpq_gettext("select() failed: %s\n"),
-							  SOCK_STRERROR(SOCK_ERRNO));
-			return EOF;
-		}
-	}
-
 	return 0;
 }
 
+/*
+ * pqReadReady: is select() saying the file is ready to read?
+ * Returns -1 on failure, 0 if not ready, 1 if ready.
+ */
+int
+pqReadReady(PGconn *conn)
+{
+	return pqSocketCheck(conn, 1, 0, (time_t) 0);
+}
+
+/*
+ * pqWriteReady: is select() saying the file is ready to write?
+ * Returns -1 on failure, 0 if not ready, 1 if ready.
+ */
+int
+pqWriteReady(PGconn *conn)
+{
+	return pqSocketCheck(conn, 0, 1, (time_t) 0);
+}
+
+/*
+ * Checks a socket, using poll or select, for data to be read, written,
+ * or both.  Returns >0 if one or more conditions are met, 0 if it timed
+ * out, -1 if an error occurred.
+ *
+ * If SSL is in use, the SSL buffer is checked prior to checking the socket
+ * for read data directly.
+ */
+static int
+pqSocketCheck(PGconn *conn, int forRead, int forWrite, time_t end_time)
+{
+	int			result;
+
+	if (!conn)
+		return -1;
+	if (conn->sock < 0)
+	{
+		printfPQExpBuffer(&conn->errorMessage,
+						  libpq_gettext("socket not open\n"));
+		return -1;
+	}
+
+#ifdef USE_SSL
+	/* Check for SSL library buffering read bytes */
+	if (forRead && conn->ssl && SSL_pending(conn->ssl) > 0)
+	{
+		/* short-circuit the select */
+		return 1;
+	}
+#endif
+
+	/* We will retry as long as we get EINTR */
+	do
+		result = pqSocketPoll(conn->sock, forRead, forWrite, end_time);
+	while (result < 0 && SOCK_ERRNO == EINTR);
+
+	if (result < 0)
+	{
+		char		sebuf[256];
+
+		printfPQExpBuffer(&conn->errorMessage,
+						  libpq_gettext("select() failed: %s\n"),
+						SOCK_STRERROR(SOCK_ERRNO, sebuf, sizeof(sebuf)));
+	}
+
+	return result;
+}
+
+
+/*
+ * Check a file descriptor for read and/or write data, possibly waiting.
+ * If neither forRead nor forWrite are set, immediately return a timeout
+ * condition (without waiting).  Return >0 if condition is met, 0
+ * if a timeout occurred, -1 if an error or interrupt occurred.
+ *
+ * Timeout is infinite if end_time is -1.  Timeout is immediate (no blocking)
+ * if end_time is 0 (or indeed, any time before now).
+ */
+static int
+pqSocketPoll(int sock, int forRead, int forWrite, time_t end_time)
+{
+	/* We use poll(2) if available, otherwise select(2) */
+#ifdef HAVE_POLL
+	struct pollfd input_fd;
+	int			timeout_ms;
+
+	if (!forRead && !forWrite)
+		return 0;
+
+	input_fd.fd = sock;
+	input_fd.events = POLLERR;
+	input_fd.revents = 0;
+
+	if (forRead)
+		input_fd.events |= POLLIN;
+	if (forWrite)
+		input_fd.events |= POLLOUT;
+
+	/* Compute appropriate timeout interval */
+	if (end_time == ((time_t) -1))
+		timeout_ms = -1;
+	else
+	{
+		time_t		now = time(NULL);
+
+		if (end_time > now)
+			timeout_ms = (end_time - now) * 1000;
+		else
+			timeout_ms = 0;
+	}
+
+	return poll(&input_fd, 1, timeout_ms);
+
+#else							/* !HAVE_POLL */
+
+	fd_set		input_mask;
+	fd_set		output_mask;
+	fd_set		except_mask;
+	struct timeval timeout;
+	struct timeval *ptr_timeout;
+
+	if (!forRead && !forWrite)
+		return 0;
+
+	FD_ZERO(&input_mask);
+	FD_ZERO(&output_mask);
+	FD_ZERO(&except_mask);
+	if (forRead)
+		FD_SET(sock, &input_mask);
+	if (forWrite)
+		FD_SET(sock, &output_mask);
+	FD_SET(sock, &except_mask);
+
+	/* Compute appropriate timeout interval */
+	if (end_time == ((time_t) -1))
+		ptr_timeout = NULL;
+	else
+	{
+		time_t		now = time(NULL);
+
+		if (end_time > now)
+			timeout.tv_sec = end_time - now;
+		else
+			timeout.tv_sec = 0;
+		timeout.tv_usec = 0;
+		ptr_timeout = &timeout;
+	}
+
+	return select(sock + 1, &input_mask, &output_mask,
+				  &except_mask, ptr_timeout);
+#endif   /* HAVE_POLL */
+}
 
 
 /*
@@ -792,7 +1086,6 @@ retry:
  * to be in fe-print.c but that file is doomed.
  */
 
-#ifdef MULTIBYTE
 /*
  * returns the byte length of the word beginning s, using the
  * specified encoding.
@@ -803,6 +1096,16 @@ PQmblen(const unsigned char *s, int encoding)
 	return (pg_encoding_mblen(encoding, s));
 }
 
+/*
+ * returns the display length of the word beginning s, using the
+ * specified encoding.
+ */
+int
+PQdsplen(const unsigned char *s, int encoding)
+{
+	return (pg_encoding_dsplen(encoding, s));
+}
+
 /*
  * Get encoding id from environment variable PGCLIENTENCODING.
  */
@@ -818,25 +1121,9 @@ PQenv2encoding(void)
 	return (encoding);
 }
 
-#else
-
-/* Provide a default definition in case someone calls it anyway */
-int
-PQmblen(const unsigned char *s, int encoding)
-{
-	(void) s;
-	(void) encoding;
-	return 1;
-}
-int
-PQenv2encoding(void)
-{
-	return 0;
-}
-#endif   /* MULTIBYTE */
-
 
 #ifdef ENABLE_NLS
+
 char *
 libpq_gettext(const char *msgid)
 {
@@ -845,52 +1132,11 @@ libpq_gettext(const char *msgid)
 	if (!already_bound)
 	{
 		already_bound = 1;
-		bindtextdomain("libpq", LOCALEDIR);
+		/* No relocatable lookup here because the binary could be anywhere */
+		bindtextdomain("libpq", getenv("PGLOCALEDIR") ? getenv("PGLOCALEDIR") : LOCALEDIR);
 	}
 
 	return dgettext("libpq", msgid);
 }
+
 #endif   /* ENABLE_NLS */
-
-#ifdef WIN32
-/*
- * strerror replacement for windows:
- *
- * This works on WIN2000 and newer, but we don't know where to find WinSock
- * error strings on older Windows flavors.  If you know, clue us in.
- */
-const char *
-winsock_strerror(int eno)
-{
-	static char	err_buf[512];
-#define WSSE_MAXLEN (sizeof(err_buf)-1-13)	/* 13 for " (0x00000000)" */
-	int			length;
-
-	/* First try the "system table", this works on Win2k and up */
-
-	if (FormatMessage(
-				FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
-					  0,
-					  eno,
-					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
-					  err_buf,
-					  WSSE_MAXLEN,
-					  NULL))
-		goto WSSE_GOODEXIT;
-
-	/* Insert other possible lookup methods here ... */
-
-	/* Everything failed, just tell the user that we don't know the desc */
-
-	strcpy(err_buf, "Socket error, no description available.");
-
-WSSE_GOODEXIT:
-
-	length = strlen(err_buf);
-	sprintf(err_buf + (length < WSSE_MAXLEN ? length : WSSE_MAXLEN),
-			" (0x%08X)", eno);
-
-	return err_buf;
-}
-
-#endif
